"""Base containers for models and their training data."""

from typing import Any, Callable, Dict, Iterable, Mapping, Tuple, Optional, Union
from abc import abstractmethod, ABC
import os

from carabiner import print_err
from datasets import Dataset
from numpy import ndarray
from pandas import DataFrame

from .aggregators import get_aggregator, AggFunction
from .data import DataMixinBase, ChemMixinBase
from .evaluation import rmse, pearson_r, spearman_r
from .information import DoubtMixinBase
from .preprocessing import Preprocessor
from .training import ModelTrainerBase
from .typing import DataLike, FeatureLike, StrOrIterableOfStr
from ..checkpoint_utils import save_json, _load_json


class ModelBoxBase(DataMixinBase, DoubtMixinBase, ABC):

    """Container class for models and their training datasets.
    
    """
    
    _prediction_key = "prediction"
    class_name = "base"
    _init_kwargs_filename = "modelbox-init-config.json"
    _special_args_filename = "modelbox-special-args.json"
    _model_config_filename = "model-config.json"

    def __init__(self, **kwargs):
        self.model = None
        self._trainer = None
        self._init_kwargs = kwargs
        self._model_config = kwargs
        self._special_args = None

    def save_checkpoint(
        self,
        checkpoint: str
    ) -> None:
        print_err(f"Saving checkpoint at {checkpoint}")
        if not os.path.exists(checkpoint):
            os.makedirs(checkpoint)
        self.save_data_checkpoint(checkpoint)
        init_kwargs = {
            "class_name": self.class_name,
        }
        init_kwargs.update(self._init_kwargs)
        save_json(init_kwargs, os.path.join(checkpoint, self.__class__._init_kwargs_filename))
        save_json(self._model_config, os.path.join(checkpoint, self.__class__._model_config_filename))
        save_json(self._special_args, os.path.join(checkpoint, self.__class__._special_args_filename))
        self.save_weights(checkpoint)
        return None

    def load_checkpoint(
        self,
        checkpoint: str,
        cache_dir: Optional[str] = None
    ) -> None:
        print_err(f"Loading checkpoint from {checkpoint}")
        self.load_data_checkpoint(checkpoint, cache_dir=cache_dir)
        self._model_config = _load_json(checkpoint, self.__class__._model_config_filename)
        self._special_args = _load_json(checkpoint, self.__class__._special_args_filename)
        if self.training_data is not None:
            self.model = self.create_model()
            self.load_weights(checkpoint, cache_dir=cache_dir)
        return None

    @abstractmethod
    def save_weights(self, checkpoint: str):
        pass

    @abstractmethod
    def load_weights(self, checkpoint: str):
        pass

    def _prepare_data(
        self, 
        data: Optional[DataLike] = None,
        features: Optional[FeatureLike] = None,
        labels: Optional[StrOrIterableOfStr] = None,
        batch_size: int = 16,
        dataloader: bool = False,
        shuffle: bool = False,
        cache: Optional[str] = None,
        dataloader_kwargs: Optional[Mapping] = None,
        **preprocessing_args
    ) -> Union[Dataset, Iterable[Mapping[str, Any]]]:  
        if data is None:
            dataset = self._check_training_data()  # from DataMixinBase
        else:
            *_, dataset = self._ingest_data(
                features=features,
                labels=labels,
                data=data,
                batch_size=batch_size,
                cache=cache,
                **preprocessing_args,
            )
        essential_keys = set([self._in_key, self._out_key])
        missing_keys = sorted(essential_keys - set(dataset.column_names))
        if len(missing_keys) > 0:
            raise KeyError(f"Dataset is missing essential columns: {', '.join(missing_keys)}.")
        if dataloader_kwargs is None:
            dataloader_kwargs = {}
        if dataloader:
            return self.make_dataloader(
                dataset=dataset, 
                batch_size=batch_size,
                shuffle=shuffle,
                **dataloader_kwargs,
            )
        else:
            return dataset

    @abstractmethod
    def create_model(self, *args, **kwargs):
        pass

    @abstractmethod
    def create_trainer(
        self, 
        callbacks: Optional[Iterable[Callable]] = None,
        epochs: int = 1, 
        *args, **kwargs,
    ) -> ModelTrainerBase:
        pass
    
    def train(
        self, 
        training_data: Optional[DataLike] = None,
        val_data: Optional[DataLike] = None,
        features: Optional[FeatureLike] = None,
        labels: Optional[StrOrIterableOfStr] = None,
        val_features: Optional[FeatureLike] = None,
        val_labels: Optional[StrOrIterableOfStr] = None,
        epochs: int = 1, 
        batch_size: int = 16,
        # learning_rate: float = .01,
        callbacks: Optional[Iterable[Callable]] = None,
        trainer_opts: Mapping[str, Union[str, int, bool]] = None,
        reinitialize: bool = False,
        **preprocessing_args
    ) -> None:
        
        """Train model with training and validation data.
    
        """
        preparation_kwargs = {
            "batch_size": batch_size,
            "dataloader": True,
        } | preprocessing_args
        training_data = self._prepare_data(
            features=features,
            labels=labels,
            data=training_data,
            shuffle=True,
            **preparation_kwargs,
        )
        if val_data is None:
            raise ValueError("No validation data provided!")
        if val_features is None:
            val_features = features
        if val_labels is None:
            val_labels = labels
        val_data = self._prepare_data(
            features=val_features,
            labels=val_labels,
            data=val_data,
            **preparation_kwargs,
        )

        if trainer_opts is None:
            trainer_opts = {}
        if self.model is None or reinitialize:
            self.model = self.create_model()
        if self._trainer is None or reinitialize:
            self._trainer = self.create_trainer(
                epochs=epochs,  # number of epochs to train for
                callbacks=callbacks,
                **trainer_opts,
            )
        self.model = self._trainer.train(
            model=self.model, 
            train_dataloader=training_data, 
            val_dataloader=val_data,
        )
        return None

    @abstractmethod
    def eval_mode(self):
        pass

    @staticmethod
    @abstractmethod
    def detach_tensor():
        pass

    @staticmethod
    def _predict(
        x: Mapping[str, Any],
        model: Callable,
        _in_key: str,
        _prediction_column: str,
        detacher_fn: Callable,
        aggregator: Optional[Callable] = None
    ) -> Dict[str, Any]:
        prediction = model(x[_in_key])
        x[_prediction_column] = detacher_fn(prediction)
        return x

    @staticmethod
    def _aggregate(
        x: Mapping[str, Any],
        _prediction_column: str,
        detacher_fn: Callable,
        aggregator: Optional[Callable] = None
    ) -> Dict[str, Any]:
        x[_prediction_column] = aggregator(detacher_fn(x[_prediction_column]))
        return x

    def predict(
        self, 
        data: Optional[DataLike] = None,
        features: Optional[FeatureLike] = None,
        labels: Optional[StrOrIterableOfStr] = None,
        batch_size: int = 16,
        aggregator: Optional[Union[str, AggFunction]] = None,
        cache: Optional[str] = None,
        agg_kwargs: Optional[Mapping] = None,
        _prediction_column: Optional[str] = None,
        **kwargs
    ) -> Dataset:

        """Make predictions on new data.
    
        """
        
        _prediction_column = (
            _prediction_column 
            or self.__class__._prediction_key
        )
        data = self._prepare_data(
            features=features,
            labels=labels,
            data=data,
            batch_size=batch_size,
            cache=cache,
            **kwargs
        )
            
        self.eval_mode()
        predictions = data.map(
            self._predict,
            fn_kwargs={
                "model": self.model,
                "detacher_fn": self.detach_tensor,
                "_in_key": self.__class__._in_key,
                "_prediction_column": _prediction_column,
            },
            batched=True, 
            batch_size=batch_size, 
            desc="Predicting",
        )
        if aggregator is not None:
            if agg_kwargs is None:
                agg_kwargs = {}
            aggregator_fn = get_aggregator(
                aggregator, 
                **agg_kwargs,
            )
            predictions = predictions.map(
                self._aggregate,
                fn_kwargs={
                    "_prediction_column": _prediction_column,
                    "aggregator": aggregator_fn,
                    "detacher_fn": self.detach_tensor,
                },
                batched=True, 
                batch_size=batch_size, 
                desc=f"Aggregating predictions by {aggregator}",
            )
        return predictions

    def evaluate(
        self, 
        data: Optional[DataLike] = None,
        features: Optional[FeatureLike] = None,
        labels: Optional[StrOrIterableOfStr] = None,
        metrics: Optional[Union[Callable, Iterable[Callable], Mapping[str, Callable]]] = None,
        batch_size: int = 16,
        aggregator: Optional[Union[str, AggFunction]] = None,
        agg_kwargs: Optional[Mapping] = None,
        cache: Optional[str] = None,
        **kwargs
    ) -> Tuple[DataFrame, Union[Tuple[float], Dict[str, float]]]:

        """Calculate metrics on training data or new data.
    
        """
        eval_prediction_col = self.__class__._prediction_key
        predictions = self.predict(
            features=features,
            labels=labels, 
            data=data, 
            batch_size=batch_size,
            aggregator=aggregator,
            agg_kwargs=agg_kwargs,
            cache=cache,
            _prediction_column=eval_prediction_col,
            **kwargs,
        )

        if metrics is None:
            metrics = {
                "rmse": rmse, 
                "pearson_r": pearson_r, 
                "spearman_rho": spearman_r,
            }
        predictions = predictions.with_format("numpy")
        y_vals = predictions[self._out_key]
        preds = predictions[eval_prediction_col]
        # if len(y_vals.shape) == 1 and len(preds.shape) == 2:
        #     y_vals = y_vals[...,None]
        print(y_vals.shape, preds.shape)
        if isinstance(metrics, Mapping):
            metrics = {
                name: metric(preds, y_vals).tolist()
                for name, metric in dict(metrics).items()
            }
        elif isinstance(metrics, (Iterable, Callable)):
            if isinstance(metrics, Callable):
                metrics = [metrics]
            metrics = tuple(
                metric(preds, y_vals).tolist()
                for metric in metrics
            )
        
        predictions = {
            key: predictions[key].flatten().tolist()
            if len(predictions[key].shape) > 1 and predictions[key].shape[-1] == 1 
            else predictions[key].tolist()
            for key in predictions.column_names
        }
        predictions = {
            key: val
            for key, val in predictions.items()
            if key != self._in_key and not key.startswith("__f__")
        }
        return DataFrame(predictions), metrics


class ModelBoxWithVarianceBase(ModelBoxBase):

    _variance_key: str = "prediction variance"
    
    def prediction_variance(
        self, 
        candidates: Optional[DataLike] = None,
        batch_size: int = 16,
        cache: Optional[str] = None,
        **kwargs
    ) -> ndarray:

        """Make predictions on new data.
    
        """
        predictions = self.predict(
            data=candidates, 
            aggregator="var", 
            agg_kwargs={"keepdims": False},
            batch_size=batch_size,
            cache=cache,
            _prediction_column=self.__class__._variance_key,
            **kwargs,
        )
        return predictions


class FingerprintModelBoxBase(ChemMixinBase, ModelBoxWithVarianceBase):

    def __init__(
        self, 
        use_fp: bool = False,
        use_2d: bool = False,
        extra_featurizers: Optional[FeatureLike] = None,
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.use_fp = use_fp
        self.use_2d = use_2d
        if extra_featurizers is not None:
            if isinstance(extra_featurizers, (str, Mapping)):
                extra_featurizers = [extra_featurizers]
        self.extra_featurizers = extra_featurizers
        self._default_preprocessing_args = {
            "structure_column": None,
            "input_representation": None,
        }
        self._model_config = kwargs

    def load_training_data(
        self,
        structure_column,
        input_representation: str = "smiles",
        features: Optional[FeatureLike] = None,
        _run_featurizer_constructor_first: bool = True,
        **kwargs
    ) -> None:
        self._default_preprocessing_args["structure_column"] = structure_column
        self._default_preprocessing_args["input_representation"] = input_representation
        if _run_featurizer_constructor_first:
            featurizer = self._featurizer_constructor(
                smiles_column=self.smiles_column,
                use_fp=self.use_fp,
                use_2d=self.use_2d,
                extra_featurizers=self.extra_featurizers,
            )
        else:
            featurizer = []
        if features is not None:
            featurizer += self._resolve_featurizers(features)
        return super().load_training_data(
            features=self._resolve_featurizers(featurizer),
            **kwargs,
            **self._default_preprocessing_args,
            smiles_column=self.smiles_column,
        )

    def train(
        self, 
        structure_column: Optional[str] = None,
        input_representation: Optional[str] = None,
        *args, **kwargs
    ) -> None:
        if structure_column is None:
            structure_column = self._default_preprocessing_args["structure_column"]
        if input_representation is None:
            input_representation = self._default_preprocessing_args["input_representation"]
        return super().train(
            *args, **kwargs, 
            structure_column=structure_column,
            input_representation=input_representation,
            smiles_column=self.smiles_column,
        )

    def predict(
        self, 
        structure_column: Optional[str] = None,
        input_representation: Optional[str] = None,
        **kwargs
    ) -> Dataset:

        """Make predictions on new data.
    
        """
        if structure_column is None:
            structure_column = self._default_preprocessing_args["structure_column"]
        if input_representation is None:
            input_representation = self._default_preprocessing_args["input_representation"]
        _extra_cols_to_keep = (
            [structure_column] + kwargs.pop("_extra_cols_to_keep", [])
        )
        return super().predict(
            **kwargs, 
            structure_column=structure_column,
            input_representation=input_representation,
            smiles_column=self.smiles_column,
            _extra_cols_to_keep=_extra_cols_to_keep,
        )

    def _info_score_entrypoint(
        self, 
        score_type: str,
        candidates,
        preprocessing_args: Optional[Mapping] = None,
        **kwargs
    ) -> Dataset:
        base_pp_args = self._default_preprocessing_args | {
            "smiles_column": self.smiles_column,
        }
        if preprocessing_args is None:
            preprocessing_args = base_pp_args
        elif isinstance(preprocessing_args, Mapping):
            preprocessing_args = base_pp_args | dict(preprocessing_args)
        else:
            raise ValueError(
                f"If provided, preprocessing_args must be a mapping, but was {type(preprocessing_args)}: {preprocessing_args}"
            )
        try:
            structure_column = preprocessing_args["structure_column"] or self._default_preprocessing_args["structure_column"]
        except KeyError:
            raise ValueError(f"You didn't supply a `structure_column`!")

        preprocessing_args["_extra_cols_to_keep"] = (
            [structure_column] + preprocessing_args.get("_extra_cols_to_keep", [])
        )
        return super()._info_score_entrypoint(
            score_type=score_type,
            candidates=candidates,
            preprocessing_args=preprocessing_args,
            **kwargs
        )


class ChempropModelBoxBase(FingerprintModelBoxBase):

    def __init__(
        self, 
        use_fp: bool = False,
        use_2d: bool = False,
        extra_featurizers: Optional[FeatureLike] = None,
        **kwargs
    ):
        super().__init__(
            use_fp=use_fp,
            use_2d=use_2d,
            extra_featurizers=extra_featurizers
        )
        self._model_config = kwargs
        self._special_args = {
            "chemprop_input_column": None,
        }

    def load_training_data(
        self,
        labels: StrOrIterableOfStr,
        features: Optional[FeatureLike] = None, 
        **kwargs
    ) -> None:
        featurizer = self._featurizer_constructor(
            smiles_column=self.smiles_column,
            use_fp=self.use_fp,
            use_2d=self.use_2d,
            extra_featurizers=self.extra_featurizers,
            _allow_no_features=True,
        )
        if features is not None:
            featurizer += features
        featurizer = self._resolve_featurizers(featurizer)
        featurizer += [{
            "name": "chemprop-mol",
            "input_column": self.smiles_column,
            "kwargs": {
                "label_column": labels,
                "extra_featurizers": [f.output_column for f in featurizer],
            },
        }]
        self._special_args["chemprop_input_column"] = Preprocessor.from_dict(featurizer[-1]).output_column
        super().load_training_data(
            labels=labels,
            features=self._resolve_featurizers(featurizer),
            _run_featurizer_constructor_first=False,
            **kwargs,
        )

    def _ingest_data(
        self, *args, **kwargs
    ):
        return super()._ingest_data(
            *args, **kwargs,
            one_column_input=self._special_args["chemprop_input_column"],
        )



"""Command-line interface for duvida."""

from typing import Mapping, Optional, Union

from argparse import FileType, Namespace
from collections import defaultdict
import os
import sys

from carabiner import cast, pprint_dict, print_err
from carabiner.cliutils import clicommand, CLIOption, CLICommand, CLIApp

from . import __version__
from .checkpoint_utils import _load_json, save_json
from .utils.package_data import _get_data_path


cache_dir, modelbox_name_file = _get_data_path("modelbox-names.json")
if not os.path.exists(modelbox_name_file):
    from .autoclass import AutoModelBox
    from .base.modelbox_registry import DEFAULT_MODELBOX, MODELBOX_NAMES
    save_json([DEFAULT_MODELBOX, MODELBOX_NAMES], modelbox_name_file)
else:
    DEFAULT_MODELBOX, MODELBOX_NAMES = _load_json(cache_dir, os.path.basename(modelbox_name_file))

_LR_DEFAULT: float = .01


@clicommand(message='Generating hyperparameter screening configurations')
def _hyperprep(args: Namespace) -> None:

    from .hyperparameters import HyperOpt

    configs = HyperOpt.from_file(args.input_file)

    for i, config in enumerate(configs):
        pprint_dict(config, message=f"Configuration {i}")

    output_dir = os.path.dirname(args.output)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    configs.write(
        args.output, 
        serialize=args.serialize,
    )
        
    return None


def _overwrite_config(
    config: Mapping, 
    config_file: Optional[str] = None, 
    config_idx: int = 0
) -> dict:

    from .hyperparameters import HyperOpt

    if config_file is not None:
        new_config = HyperOpt.from_file(config_file, silent=True)._ranges[config_idx] 
        pprint_dict(
            new_config,
            message=f"Overriding command-line parameters from config file {config_file}",
        )
        # new_config.update({key: val in config.items() if val is not None})  # command line takes precedent
        config.update(new_config)  # config takes precedent
        pprint_dict(
            config,
            message="Initialization parameters are now",
        )
        return config
    else:
        return config


def _init_modelbox(
    cli_config: Mapping[str, Union[str, int, float]],
    checkpoint: Optional[str] = None,
    config_file: Optional[str] = None,
    config_idx: int = 0,
    cache: Optional[str] = None,
    **overrides
):
    from .autoclass import AutoModelBox
    if checkpoint is None:
        if any(
            overrides.get(key) is None for key in ("training", "labels")
        ) and all(
            overrides.get(key) is None
            for key in ("structure", "features")
        ):
                raise ValueError(
                    """If not providing a checkpoint, --training and --labels, 
                    and either --features or --structure must be set.
                    """
                )
        cli_config = _overwrite_config(
            cli_config, 
            config_file=config_file, 
            config_idx=config_idx,
        )
        modelbox = AutoModelBox(**cli_config)._instance
    else:
        modelbox = AutoModelBox.from_pretrained(checkpoint, cache=cache)
    return modelbox
    

def _load_modelbox_training_data(
    modelbox,
    checkpoint: Optional = None,
    cache: Optional[str] = None,
    **overrides
):
    if any([
        overrides.get("training") is not None,  # override checkpoint training data
        checkpoint is None,  # no checkpoint
        modelbox.training_data is None,  # checkpoint without training data
    ]):
        load_data_args = {
            "data": overrides.get("training"), 
            "cache": cache,
            # command-line takes precedent:
            "features": overrides.get("features") or modelbox._input_featurizers,
            "labels": overrides.get("labels") or modelbox._label_cols, 
        }
        if hasattr(modelbox, "tanimoto_column"):  # i.e., is for chemistry
            # command-line takes precedent:
            load_data_args["structure_column"] = overrides.get("structure") or modelbox.structure_column
        pprint_dict(
            load_data_args,
            message="Data-loading configuration",
        )
        modelbox.load_training_data(**load_data_args)
    return modelbox, load_data_args


def _init_modelbox_and_load_training_data(
    cli_config: Mapping[str, Union[str, int, float]],
    checkpoint: Optional[str] = None,
    config_file: Optional[str] = None,
    config_idx: int = 0,
    cache: Optional[str] = None,
    **overrides
):
    modelbox = _init_modelbox(
        cli_config=cli_config,
        checkpoint=checkpoint,
        config_file=config_file,
        config_idx=config_idx,
        cache=cache,
        **overrides,
    )

    modelbox, load_data_args = _load_modelbox_training_data(
        modelbox=modelbox,
        checkpoint=checkpoint,
        cache=cache,
        **overrides,
    )

    if checkpoint is None:  # model not instantiated yet
        modelbox.model = modelbox.create_model()
    return modelbox, load_data_args


def _train_and_save_modelbox(
    modelbox,
    early_stopping: Optional[int] = None,
    output_name: Optional[str] = None,
    **training_args
):
    from datasets.fingerprint import Hasher
    from lightning.pytorch.callbacks import EarlyStopping
    from lightning.pytorch.loggers import CSVLogger, TensorBoardLogger

    from .utils.lightning import _get_most_recent_lightning_log
    from .utils.plotting import _plot_history

    if early_stopping is not None:
        callbacks = [EarlyStopping('val_loss', patience=early_stopping)]
    else:
        callbacks = None

    if output_name is None:
        output_name = (
            f"{modelbox.class_name}_n{modelbox.size}_"
            f"y={'-'.join(modelbox._label_cols)}_"
            f"h={Hasher.hash(modelbox._input_featurizers)}",
        )
    
    checkpoint_path = output_name
    modelbox.train(
        callbacks=callbacks,
        trainer_opts={  # passed to lightning.Trainer()
            "logger": [
                CSVLogger(save_dir=os.path.join(checkpoint_path, "logs-csv")),
                TensorBoardLogger(save_dir=os.path.join(checkpoint_path, "logs")),
            ], 
            "enable_progress_bar": True, 
            "enable_model_summary": True,
        },
        **training_args,
    )

    # get latest CSV log
    max_version = _get_most_recent_lightning_log(
        os.path.join(checkpoint_path, "logs-csv"),
        "metrics.csv",
    )
    _plot_history(
        max_version,
        os.path.join(checkpoint_path, "training-log")
    )
    modelbox.save_checkpoint(checkpoint_path)
        
    return checkpoint_path, training_args


def _evaluate_modelbox_and_save_metrics(
    modelbox,
    metric_filename: str,
    plot_filename: str,
    dataset: Optional = None,
    **kwargs
):
    import torch
    from .utils.plotting import _plot_prediction_scatter

    modelbox.to("cuda" if torch.cuda.is_available() else "cpu")
    predictions, metrics = modelbox.evaluate(
        data=dataset,
        aggregator="mean",
        agg_kwargs={"keepdims": True},
        **kwargs,
    )
    save_json(
        metrics, 
        metric_filename,
    )
    _plot_prediction_scatter(
        predictions,
        x=modelbox._prediction_key,
        y=modelbox._out_key,
        filename=plot_filename,
    )
    return metrics | {
        "model_class": modelbox.class_name,
        "n_parameters": modelbox.size,
    }


def _dict_to_pandas(
    d: Mapping,
    filename: Optional[str] = None
):
    import pandas as pd
    try:
        df = pd.DataFrame(d)
    except ValueError as e:  # not all columns same length; should never happen
        pprint_dict(
            {key: len(val) for key, val in d.items()},
            message="Metrics table column lengths"
        )
        raise e
    if filename is not None:
        df.to_csv(filename, index=False)
    return df


@clicommand(message='Training a Pytorch model')
def _train(args: Namespace) -> None:

    from .autoclass import AutoModelBox

    cli_config = {
        "class_name": args.model_class.casefold(),
        "use_2d": args.descriptors,
        "use_fp": args.fp,
        "n_hidden": args.hidden,
        "residual_depth": args.residual,
        "n_units": args.units,
        "dropout": args.dropout,
        "ensemble_size": args.ensemble_size,
        "learning_rate": args.learning_rate,
    }

    modelbox, load_data_args = _init_modelbox_and_load_training_data(
        cli_config=cli_config,
        checkpoint=args.checkpoint,
        config_file=args.config,
        config_idx=args.config_index,
        cache=args.cache,
        # overrides:
        training=args.training,
        structure=args.structure,
        structure_representation=args.input_representation,
        labels=args.labels,
        features=args.features,
    )

    pprint_dict(
        modelbox._model_config, 
        message=f"Initialized model {modelbox.class_name} with {modelbox.size} parameters",
    )
    
    training_args = {
        "epochs": args.epochs, 
        "batch_size": args.batch,
        "val_data": args.validation,
        "early_stopping": args.early_stopping,
    }
    pprint_dict(
        training_args, 
        message=f">> Training {modelbox.class_name} with training configuration",
    )
    checkpoint_path, training_args = _train_and_save_modelbox(
        modelbox=modelbox,
        early_stopping=args.early_stopping,
        epochs=args.epochs, 
        batch_size=args.batch,
        val_data=args.validation,
        output_name=args.output,
    )
    for obj, f in zip((training_args, load_data_args), ("training-args.json", "load-data-args.json")):
        save_json(obj, os.path.join(checkpoint_path, f))

    # Reload - built-in test that the checkpointing works!
    modelbox = AutoModelBox.from_pretrained(
        checkpoint_path, 
        cache=args.cache,
    )
    overall_metrics = defaultdict(list)
    for name in ("training", "validation", "test"):
        dataset = getattr(args, name)
        if dataset is not None:  # skip optional extra datasets, e.g. "test"
            if name == "training":
                dataset = None  # Use cached training data
            metrics = _evaluate_modelbox_and_save_metrics(
                modelbox,
                metric_filename=os.path.join(checkpoint_path, f"eval-metrics_{name}.json"),
                plot_filename=os.path.join(checkpoint_path, f"predictions_{name}"),
                dataset=dataset,
            )
            pprint_dict(
                metrics, 
                message=f"Evaluation: {name}",
            )

            overall_metrics["split"].append(name)
            overall_metrics["split_filename"].append(dataset or load_data_args["data"])
            if args.config is not None:
                overall_metrics["config_i"].append(args.config_index)
            keys_added = []
            for d in (
                modelbox._init_kwargs, 
                modelbox._model_config, 
                load_data_args, 
                training_args, 
                metrics,
            ):
                for key, val in d.items():
                    if key != "trainer_opts" and key not in keys_added:
                        overall_metrics[key].append(val)
                        keys_added.append(key)

    _dict_to_pandas(overall_metrics, os.path.join(checkpoint_path, "metrics.csv"))

    return None


def _resolve_and_slice_data(
    data: str,
    start: Optional[int] = None,
    end: Optional[int] = None,
    batch_size: int = 1024
):
    from .base.data import DataMixinBase
    # from .utils.datasets import to_dataset

    candidates_ds = DataMixinBase._resolve_data(data)
    nrows = candidates_ds.num_rows
    skip = start or 0
    take = (end or nrows) - skip
    if (take - skip) < nrows:
        print_err(f"INFO: Reading dataset from row {skip} to row {take + skip} / {nrows}.")
    return candidates_ds.skip(skip).take(take)


def _save_dataset(
    dataset,
    output: str
) -> None:
    print_err("INFO: Saving dataset:\n" + str(dataset) + "\n" + f"at {output} as", end=" ")
    if output.endswith((".csv", ".csv.gz", ".tsv", ".tsv.gz", ".txt", ".txt.gz")):
        print_err("CSV.")
        dataset.to_csv(
            output, 
            sep="," if output.endswith((".csv", ".csv.gz")) else "\t",
            compression='gzip' if output.endswith(".gz") else None,
        )
    elif output.endswith(".json"):
        print_err("JSON.")
        dataset.to_json(output)
    elif output.endswith(".parquet"):
        print_err("Parquet.")
        dataset.to_parquet(output)
    elif output.endswith(".sql"):
        print_err("SQL.")
        dataset.to_sql(output)
    elif output.endswith(".hf"):
        print_err("Hugging Face dataset.")
        dataset.save_to_disk(output)
    else:
        print_err("Hugging Face dataset.")
        dataset.save_to_disk(output + ".hf")
        print_err(f"WARNING: Unsure what format to save as for filename {output}. Defaulted to Hugging Face dataset.")
    return None


@clicommand("Predicting with the following parameters")
def _predict(args: Namespace) -> None:

    import torch
    from .autoclass import AutoModelBox

    output = args.output
    out_dir = os.path.dirname(output)
    if len(out_dir) > 0 and not os.path.exists(out_dir):
        os.makedirs(out_dir)
    preprocessing_args = {
        "structure_column": args.structure,
        "input_representation": args.input_representation,
    }
    common_args = {
        "batch_size": args.batch,
        "cache": args.cache,
    }
    candidates_ds = _resolve_and_slice_data(
        args.test,
        start=args.start,
        end=args.end,
    )
    modelbox = AutoModelBox.from_pretrained(
        args.checkpoint, 
        cache=args.cache,
    )
    pprint_dict(
        modelbox._model_config, 
        message=f"Initialized model {modelbox.class_name} with {modelbox.size} parameters",
    )
    for col in cast(modelbox._label_cols, to=list):
        if col not in candidates_ds.column_names:
            from numpy import zeros_like
            candidates_ds = candidates_ds.add_column(
                col,
                zeros_like(
                    candidates_ds
                    .with_format("numpy")
                    [candidates_ds.column_names[0]]
                )
            )
    
    preprocessing_args["_extra_cols_to_keep"] = (args.extras or [])
    modelbox.to("cuda" if torch.cuda.is_available() else "cpu")
    candidates_ds = modelbox.predict(
        data=candidates_ds,
        aggregator="mean",
        features=args.features,
        **preprocessing_args,
        **common_args,
    )
    preprocessing_args["_extra_cols_to_keep"].append(modelbox._prediction_key)
    if args.variance:
        candidates_ds = modelbox.prediction_variance(
            candidates=candidates_ds,
            features=args.features,
            **preprocessing_args,
            **common_args,
        )
        preprocessing_args["_extra_cols_to_keep"].append(modelbox._variance_key)
    if args.tanimoto:
        if hasattr(modelbox, "tanimoto_nn"):
            candidates_ds = modelbox.tanimoto_nn(
                data=candidates_ds,
                query_structure_column=args.structure,
                query_input_representation=args.input_representation,
                **common_args,
            )
            preprocessing_args["_extra_cols_to_keep"].append(modelbox.tanimoto_column)
        else:
            print_err(f"Cannot calculate Tanimoto for non-chemical modelbox from {args.checkpoint}")
    if args.doubtscore:
        modelbox.model.set_model(0)
        candidates_ds = modelbox.doubtscore(
            candidates=candidates_ds,
            features=args.features,
            preprocessing_args=preprocessing_args,
            **common_args,
        )
        preprocessing_args["_extra_cols_to_keep"].append("doubtscore")
    if args.information_sensitivity:
        modelbox.model.set_model(0)
        if args.approx == "bekas":
            extra_args = {"n": args.bekas_n}
        else:
            extra_args = {}
        candidates_ds = modelbox.information_sensitivity(
            candidates=candidates_ds,
            features=args.features,
            preprocessing_args=preprocessing_args,
            approximator=args.approx,
            optimality_approximation=args.optimality,
            **common_args,
            **extra_args,
        )
        preprocessing_args["_extra_cols_to_keep"].append("information sensitivity")
        
    print_err(preprocessing_args)

    _save_dataset(
        candidates_ds.remove_columns([modelbox._in_key, modelbox._out_key]), 
        output,
    )

    if args.labels is not None:
        overall_metrics = defaultdict(list)
        metric_filename = os.path.join(out_dir, "predict-eval-metrics-table.csv")
        plot_filename = os.path.join(out_dir, "predict-eval-scatter")
        metrics = _evaluate_modelbox_and_save_metrics(
            modelbox,
            dataset=candidates_ds,
            **preprocessing_args,
            metric_filename=metric_filename,
            plot_filename=plot_filename,
        )
        pprint_dict(
            metrics, 
            message=f"Evaluation",
        )
        overall_metrics["model_class"].append(modelbox.class_name)
        overall_metrics["n_parameters"].append(modelbox.size)
        keys_added = set(overall_metrics.keys())
        for d in (
            modelbox._init_kwargs, 
            modelbox._model_config, 
            metrics,
        ):
            for key, val in d.items():
                if key != "trainer_opts" and key not in keys_added:
                    overall_metrics[key].append(val)
                    keys_added.add(key)
        _dict_to_pandas(
            overall_metrics, 
            os.path.join(out_dir, "metrics.csv"),
        )

    return None


@clicommand("Splitting data with the following parameters")
def _split(args: Namespace) -> None:

    from .utils.splitting import split_dataset
    output = args.output
    out_dir = os.path.dirname(output)
    if len(out_dir) > 0 and not os.path.exists(out_dir):
        os.makedirs(out_dir)
    
    if args.train is None:
        raise ValueError(f"You need to at least provide a --train fraction.")
    
    ds = _resolve_and_slice_data(
        args.input_file,
        start=args.start,
        end=args.end,
    )
    if args.type == "faiss":
        faiss_opts = {
            "cache": args.cache,
            "n_neighbors": args.n_neighbors,
        }
    else:
        faiss_opts = {}
    ds, splits = split_dataset(
        ds=ds,
        method=args.type,
        structure_column=args.structure,
        input_representation=args.input_representation,
        train=args.train,
        validation=args.validation,
        test=args.test,
        batch_size=args.batch,
        seed=args.seed or 42,
        deterministic=args.seed is not None,
        **faiss_opts,
    )
    root, ext = os.path.splitext(output)
    for key, split_ds in splits.items():
        _save_dataset(
            split_ds, 
            f"{root}_{key}{ext}",
        )

    if args.plot is not None:

        from carabiner.mpl import figsaver
        from .utils.splitting.plot import plot_chemical_splits

        print_err(f"Plotting splits...")
        
        (fig, axes), df = plot_chemical_splits(
            ds=ds,
            structure_column=args.structure,
            input_representation=args.input_representation,
            split_columns="split",
            sample_size=args.plot_sample,
            additional_columns=args.extras,
            seed=args.plot_seed,
            cache=args.cache,
        )
        root, ext = os.path.splitext(args.plot)
        figsaver(format=ext.lstrip("."))(fig, root, df=df)
    return None


@clicommand("Tagging data percentiles with the following parameters")
def _percentile(args: Namespace) -> None:

    if args.plot is not None and args.structure is None:
        raise ValueError(
            f"""
            If you want to save a plot at "{args.plot}", then you need to provide
            a chemical structure (like SMILES) column name using --structure so 
            that a UMAP embedding can be calculated.
            """
        )

    from .utils.splitting.top_k import percentiles
    
    output = args.output
    out_dir = os.path.dirname(output)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    
    ds = _resolve_and_slice_data(
        args.input_file,
        start=args.start,
        end=args.end,
    )
    print(ds)
    # for b in ds:
    #     print(b)
    #     break
    q = {col: args.percentiles for col in args.columns}
    ds = percentiles(
        ds=ds,
        q=q,
        compression=args.compression,
        delta=args.delta,
        reverse=args.reverse,
        cache=args.cache,
    )
    _save_dataset(ds, output)

    if args.plot is not None:

        from carabiner.mpl import figsaver
        from .utils.splitting.plot import plot_chemical_splits

        print_err(f"Plotting top percentiles...")

        (fig, axes), df = plot_chemical_splits(
            ds=ds,
            structure_column=args.structure,
            input_representation=args.input_representation,
            split_columns=[col for col in ds.column_names if any(f"{_q}_top_" in col for _q in q)],
            sample_size=args.plot_sample,
            additional_columns=args.extras,
            seed=args.plot_seed,
            cache=args.cache,
        )
        root, ext = os.path.splitext(args.plot)
        figsaver(format=ext.lstrip("."))(fig, root, df=df)
    return None


def main() -> None:

    input_file = CLIOption(
        'input_file',
        type=FileType('r'),
        default=sys.stdin,
        nargs='?',
        help='Input file. Default: STDIN',
    )
    input_filename = CLIOption(
        'input_file',
        type=str,
        help='Input file.',
    )
    train_data = CLIOption(
        '--training', '-1',
        type=str,
        default=None,
        help='Training dataset. Required if no checkpoint provided.',
    )
    val_data = CLIOption(
        '--validation', '-2',
        type=str,
        required=True,
        help='Validation dataset file.',
    )
    test_data = CLIOption(
        '--test', '-t',
        type=str,
        default=None,
        help='Test dataset file.',
    )
    feature_cols = CLIOption(
        '--features', '-x',
        type=str,
        nargs='*',
        default=None,
        help='Column names from data file that contain features. Required if no checkpoint provided.',
    )
    structure_col = CLIOption(
        '--structure', '-S',
        type=str,
        default=None,
        help="""
        Column names from data file that contains a string representation 
        of chemical structure. Required for chemical model classes.
        """,
    )
    structure_representation = CLIOption(
        '--input-representation', '-R',
        type=str,
        default="smiles",
        choices=["smiles", "selfies", "inchi", "aa_seq"],
        help='Type of chemical structure string. Default: SMILES if training; for prediction, use same as training data.',
    )
    label_cols = CLIOption(
        '--labels', '-y',
        type=str,
        nargs='*',
        default=None,
        help='Column names from data file that contain labels. Required if no checkpoint provided.',
    )
    cache = CLIOption(
        '--cache',
        type=str,
        default=".",
        help='Where to cache data.',
    )
    model_class = CLIOption(
        '--model-class', '-k',
        type=str,
        default=DEFAULT_MODELBOX,
        choices=MODELBOX_NAMES,
        help='Test dataset file.',
    )
    _checkpoint = CLIOption(
        '--checkpoint',
        type=str,
        default=None,
        help='Load a modelbox from this checkpoint. Default: do not use, make a new modelbox.',
    )
    n_units = CLIOption(
        '--units', '-u',
        type=int,
        default=8,
        help='Number of units per hidden layer.',
    )
    n_hidden = CLIOption(
        '--hidden', '-m',
        type=int,
        default=1,
        help='Number of hidden layers.',
    )
    residual_depth = CLIOption(
        '--residual',
        type=int,
        default=None,
        help='Depth of residual blocks. Default: Do not use residual blocks.',
    )
    _2d = CLIOption(
        '--descriptors', 
        action="store_true",
        help='Use 2d descriptors (needs a SMILES input feature).',
    )
    _fp = CLIOption(
        '--fp', 
        action="store_true",
        help='Use chemical fingerprints (needs a SMILES input feature).',
    )
    dropout = CLIOption(
        '--dropout', '-d',
        type=float,
        default=0.,
        help='Dropout rate for training.',
    )
    ensemble_size = CLIOption(
        '--ensemble-size', '-z',
        type=int,
        default=1,
        help='Number of models to train in an ensemble.',
    )
    batch_size = CLIOption(
        '--batch', '-b',
        type=int,
        default=16,
        help='Batch size for training.',
    )
    n_epochs = CLIOption(
        '--epochs', '-e',
        type=int,
        default=1,
        help='Number of epochs for training.',
    )
    learning_rate = CLIOption(
        '--learning-rate', '-r',
        type=float,
        default=None,
        help=f'Learning rate for training. Default: {_LR_DEFAULT}.',
    )
    early_stopping = CLIOption(
        '--early-stopping', '-s',
        type=int,
        default=None,
        help='Number of epochs to wait for improvement before stopping. Default: no early stopping.',
    )
    model_config = CLIOption(
        '--config', '-c',
        type=str,
        default=None,
        help='Model configuration file. Overrides other options.',
    )
    config_i = CLIOption(
        '--config-index', '-i',
        type=int,
        default=0,
        help='If more than one config in `--config`, choose this one.',
    )
    output_name = CLIOption(
        '--output', '-o', 
        type=str,
        required=True,
        help='Output filename.',
    )
    serialize = CLIOption(
        '--serialize', '-z', 
        action="store_true",
        help='Pickle instead of JSON output.',
    )
    # output = CLIOption(
    #     '--output', '-o', 
    #     type=FileType('w'),
    #     default=sys.stdout,
    #     help='Output file. Default: STDOUT',
    # )
    # formatting = CLIOption(
    #     '--format', '-f', 
    #     type=str,
    #     default='TSV',
    #     choices=['TSV', 'CSV', 'tsv', 'csv'],
    #     help='Format of files. Default: %(default)s',
    # )

    # slice dataset
    slice_start = CLIOption(
        '--start', 
        type=int,
        default=0,
        help='First row of dataset to process.',
    )
    slice_end = CLIOption(
        '--end', 
        type=int,
        default=None,
        help='Last row of dataset to process. Default: end of dataset.',
    )
    extra_cols = CLIOption(
        '--extras',
        type=str,
        nargs="*",
        default=None,
        help='Extra columns to retain in prediction table; useful for IDs.',
    )

    # Information metrics
    variance = CLIOption(
        '--variance', 
        action="store_true",
        help='Calculate ensemble variance.',
    )
    tanimoto = CLIOption(
        '--tanimoto', 
        action="store_true",
        help='Calculate Tanimoto distance to nearest neighbor in training data.',
    )
    doubtscore = CLIOption(
        '--doubtscore', 
        action="store_true",
        help='Calculate doubtscore.',
    )
    info_sens = CLIOption(
        '--information-sensitivity', 
        action="store_true",
        help='Calculate information senstivity.',
    )
    optimality = CLIOption(
        '--optimality', 
        action="store_true",
        help='Whether to make the computationally faster assumption that the model parameters were trained to gradient 0.',
    )
    hess_approx = CLIOption(
        '--approx', 
        type=str,
        default="bekas",
        choices=["exact", "squared_jacobian", "rough_finite_difference", "bekas"],
        help='What type of Hessian approximation to perform.',
    )
    bekas_n = CLIOption(
        '--bekas-n', 
        type=int,
        default=1,
        help='Number of stochastic samples for Hessian approximation.',
    )

    split_type = CLIOption(
        '--type', 
        type=str,
        default="scaffold",
        choices=["scaffold", "faiss"],
        help='Splitting method.',
    )
    random_seed = CLIOption(
        '--seed', '-i', 
        type=int,
        default=None,
        help='Random seed. Default: determininstic.',
    )
    n_neighbors = CLIOption(
        '--n-neighbors', '-k', 
        type=int,
        default=10,
        help='Number of nearest neighbors for FAISS splitting.',
    )
    train_test_val = [
        CLIOption(
            f'--{key}', 
            type=float,
            default=None,
            help='Fraction of examples for each split. Default: infer.',
        ) for key in ("train", "validation", "test")
    ]

    columns = CLIOption(
        '--columns', '-c',
        type=str,
        nargs='*',
        help='List of columns to tag percentiles for.',
    )
    percentiles = CLIOption(
        '--percentiles', '-p', 
        type=float,
        nargs='*',
        default=[5.],
        help='List of percentiles to calculate.',
    )
    reverse = CLIOption(
        '--reverse', '-r', 
        action='store_true',
        help='Whether to reverse percentiles (i.e. high to low).',
    )
    do_plot = CLIOption(
        '--plot', 
        type=str,
        default=None,
        help='Filename to save UMAP plot under.',
    )
    compression = CLIOption(
        '--compression', '-z', 
        type=int,
        default=500,
        help='How many centroids for quantile approximation. Higher is more accurate, but uses more memory.',
    )
    delta = CLIOption(
        '--delta', '-d', 
        type=float,
        default=1.,
        help='Width from percentile cutoff to buffer as borderline for refinement. Higher is more accurate, but uses more memory.',
    )
    plot_seed = CLIOption(
        '--plot-seed', '-e', 
        type=int,
        default=42,
        help='Seed for UMAP embedding.',
    )
    plot_sample = CLIOption(
        '--plot-sample', '-n', 
        type=int,
        default=20_000,
        help='Subsample size for UMAP embedding.',
    )
    extras = CLIOption(
        '--extras', '-x',
        type=str,
        nargs='*',
        help='Additional columns for coloring UMAP plot.',
    )

    hyperprep = CLICommand(
        "hyperprep",
        description="Prepare inputs for hyperparameter search.",
        options=[
            input_file, 
            serialize,
            output_name,
        ],
        main=_hyperprep,
    )

    train = CLICommand(
        "train",
        description="Train a PyTorch model.",
        options=[
            train_data, 
            val_data,
            test_data,
            feature_cols,
            structure_col,
            structure_representation,
            label_cols,
            model_class,
            _checkpoint,
            n_units,
            n_hidden,
            residual_depth,
            _2d, 
            _fp,
            dropout,
            batch_size,
            n_epochs,
            learning_rate,
            early_stopping,
            ensemble_size,
            model_config,
            config_i,
            output_name,
            cache,
        ],
        main=_train,
    )

    predict = CLICommand(
        "predict",
        description="Make predictions and calculate uncertainty using a duvida checkpoint.",
        options=[
            test_data, 
            slice_start,
            slice_end,
            feature_cols,
            label_cols,
            structure_col,
            extra_cols,
            structure_representation,
            _checkpoint,
            cache,
            output_name,
            variance,
            tanimoto,
            doubtscore,
            info_sens,
            optimality,
            hess_approx,
            bekas_n,
            batch_size,
        ],
        main=_predict,
    )

    split = CLICommand(
        "split",
        description="Make chemical train-test-val splits on out-of-core datasets.",
        options=[
            input_filename, 
            split_type,
            n_neighbors,
            slice_start,
            slice_end,
            structure_col,
            structure_representation,
            do_plot,
            plot_sample,
            plot_seed,
            extras,
            random_seed,
            cache,
            output_name,
            batch_size,
        ] + train_test_val,
        main=_split,
    )

    percentiles = CLICommand(
        "percentiles",
        description="Add columns indicating whether rows are in a percentile.",
        options=[
            input_filename, 
            columns,
            percentiles,
            reverse,
            compression,
            delta,
            slice_start,
            slice_end,
            cache,
            output_name,
            batch_size,
            do_plot,
            structure_col,
            structure_representation,
            plot_sample,
            plot_seed,
            extras,
        ],
        main=_percentile,
    )

    app = CLIApp(
        "duvida", 
        version=__version__,
        description=(
            "Calculating exact and approximate confidence and "
            "information metrics for deep learning on general "
            "purpose and chemistry tasks."
        ),
        commands=[
            hyperprep,
            train,
            predict,
            split,
            percentiles,
        ],
    )

    app.run()
    return None


if __name__ == '__main__':
    main()



"""Enabling information sensitivity calculations."""

from typing import Callable, Dict, Mapping, Tuple, Optional, Union
from functools import partial

import torch
from torch.func import functional_call, replace_all_batch_norm_modules_
from torch.nn import Module, Parameter

from ...base.information import DoubtMixinBase
from ...stateless.config import config

config.set_backend('torch', precision='float')

from ...stateless.hessians import _DEFAULT_APPROXIMATOR
from ...stateless.information import (
    fisher_score, 
    fisher_information_diagonal, 
    parameter_gradient, 
    parameter_hessian_diagonal,
    parameter_gradient_unrolled, 
    parameter_hessian_diagonal_unrolled
)
from ...stateless.typing import Array, ArrayLike, LossFunction, StatelessModel
from ...stateless.utils import get_eps, jit, vmap
from ..functions import mse_loss
from ..models.utils.ensemble import TorchEnsembleMixin

_EPS = get_eps()
_DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class DoubtMixin(DoubtMixinBase):

    device = _DEVICE

    @staticmethod
    def _get_shape(x: ArrayLike) -> Array:
        try:
            return x.shape
        except AttributeError:
            return torch.ones((1,))

    def to(self, device: str) -> None:
        device = torch.device(device)
        state_dict = self.model.state_dict()
        parameter_names = {name for name, param in self.model.named_parameters()}
        new_state_dict = {
            key: val.to(device)
            for key, val in state_dict.items()
            if key in parameter_names
        }
        new_state_dict.update({
            key: Parameter(
                val.to(device), 
                requires_grad=False,  # otherwise batchnorm integers cause error (not differentiable)
            ) 
            for key, val in state_dict.items()
            if key not in parameter_names
        })
        self.model.load_state_dict(
            new_state_dict, 
            assign=True,  # otherwise new device won't be preserved
        )
        self.device = device
        return None    

    @staticmethod
    def release_memory() -> None:
        torch.cuda.empty_cache()
        return None

    @staticmethod
    def _pack_params_like(
        p: ArrayLike, 
        d: Mapping[str, ArrayLike]
    ) -> Dict[str, Array]:
        index = 0
        output = {}
        for name, param in d.items():
            end_index = index + param.numel()
            output[name] = p[index:end_index].view(-1, *param.shape)
            if output[name].shape[0] == 1:
                output[name] = output[name].squeeze(0)
            index = end_index
        return output

    def make_stateless_model(
        self, 
        model: Optional[Module] = None,
        flat_params: bool = False
    ) -> Union[Tuple[StatelessModel, Dict[str, Array]], Tuple[StatelessModel, Dict[str, Array], Array]]:
        if model is None:
            self.to(self.device)
            model = self.model
        else:
            model.to(self.device)
        model.eval()
        replace_all_batch_norm_modules_(model)
        model_params = {key: val.detach().clone() for key, val in model.named_parameters()}
        if isinstance(model, TorchEnsembleMixin):
            model_params = {
                key: val for key, val in model_params.items()
                if any(model_key in key for model_key in model.model_keys)
            }

        def stateless_model(x, *params):
            # x = x.to(self.device)
            return functional_call(
                module=model, 
                parameter_and_buffer_dicts=params, 
                args=(x,),
                tie_weights=False,
            )

        if flat_params:

            # @jit  # Fails compile
            def stateless_model_flat_params(x, *flat_params):
                try:
                    flat_params = torch.concat(flat_params)
                except RuntimeError:  # if flat_params has already gone through *tuple unpacking
                    flat_params = torch.stack(flat_params)
                packed_params = self._pack_params_like(
                    torch.as_tensor(flat_params), 
                    model_params,
                )
                return stateless_model(x, packed_params)
            
            flattened_params = torch.concat([p.flatten() for _, p in model_params.items()])
            return stateless_model_flat_params, model_params, flattened_params
        else:
            return stateless_model, model_params

    def parameter_gradient(
        self,
        model: Optional[StatelessModel] = None
    ) -> Callable[[ArrayLike], Array]:
        stateless_model, params = self.make_stateless_model(model)
        gradient_fn = parameter_gradient(stateless_model)

        @jit
        def _parameter_gradient_with_params(x: ArrayLike) -> Array:
            return gradient_fn((params,), x)[0]

        return _parameter_gradient_with_params

    def fisher_score(
        self, 
        model: Optional[StatelessModel] = None,
        loss: LossFunction = mse_loss,
    ) -> Callable[[ArrayLike, ArrayLike], Array]:
        stateless_model, params = self.make_stateless_model(model)
        gradient_fn = fisher_score(stateless_model, loss)
        
        @jit
        def _fisher_score_with_params(x_true: ArrayLike, y_true: ArrayLike) -> Array:
            return gradient_fn((params,), x_true, y_true.to(self.device))[0]
        
        return _fisher_score_with_params
    
    def parameter_hessian_diagonal(
        self, 
        model: Optional[StatelessModel] = None,
        approximator: str = _DEFAULT_APPROXIMATOR, 
        *args, **kwargs
    ) -> Callable:
        stateless_model, params, flat_params = self.make_stateless_model(model, flat_params=True)
        if approximator in ('bekas', 'exact_diagonal'):
            kwargs['device'] = self.device
        hessian_fn = parameter_hessian_diagonal(
            stateless_model,
            approximator=approximator,
            *args, **kwargs,
        )
        packing_fn = partial(vmap, in_axes=(0, None))(self._pack_params_like)

        @jit
        def _parameter_hessian_with_params(x):
            return packing_fn(hessian_fn(flat_params, x), params)

        return _parameter_hessian_with_params
    
    def fisher_information_diagonal(
        self, 
        model: Optional[StatelessModel] = None,
        loss: LossFunction = mse_loss,
        approximator: str = _DEFAULT_APPROXIMATOR, 
        *args, **kwargs
    ) -> Callable:
        stateless_model, params, flat_params = self.make_stateless_model(model, flat_params=True)
        if approximator in ('bekas', 'exact_diagonal'):
            kwargs['device'] = self.device
        fisher_info_fn = fisher_information_diagonal(
            stateless_model,
            loss,
            approximator=approximator,
            *args, **kwargs,
        )

        # @jit  # Doesn't compile
        def _fisher_inform_diagonal_with_params(x_true, y_true):
            return self._pack_params_like(fisher_info_fn(flat_params, x_true, y_true.to(self.device)), params)

        return _fisher_inform_diagonal_with_params

    @jit
    def doubtscore_core(
        self,
        fisher_score: ArrayLike, 
        parameter_gradient: ArrayLike, 
        eps: float = _EPS
    ) -> Array:
        doubtscore = parameter_gradient / (fisher_score.unsqueeze(0).to(self.device) + eps)
        return doubtscore.flatten(start_dim=1, end_dim=-1).detach().cpu()
    
    @jit
    def information_sensitivity_core(
        self, 
        fisher_score: ArrayLike, 
        fisher_information_diagonal: ArrayLike, 
        parameter_gradient: ArrayLike, 
        parameter_hessian_diagonal: Optional[ArrayLike] = None, 
        eps: float = _EPS,
        optimality_approximation: bool = False
    ) -> Array:
        term1 = fisher_information_diagonal.unsqueeze(0).to(self.device) * parameter_gradient
        if not (optimality_approximation or parameter_hessian_diagonal is None):
            term2 = fisher_score.unsqueeze(0).to(self.device) * parameter_hessian_diagonal
        else:
            term2 = 0.
        information_sensitivity = torch.square(parameter_gradient) / (term1 - term2 + eps)
        return information_sensitivity.flatten(start_dim=1, end_dim=-1).detach().cpu()


class ChempropDoubtMixin(DoubtMixin):

    def parameter_gradient(
        self,
        model: Optional[StatelessModel] = None
    ) -> Callable[[ArrayLike], Array]:
        stateless_model, params = self.make_stateless_model(model)
        gradient_fn = parameter_gradient_unrolled(stateless_model)

        # @jit  # fails compile
        def _parameter_gradient_with_params(x: ArrayLike) -> Array:
            g = [_g[0] for _g in gradient_fn((params,), x)]
            return {
                key: torch.stack([_g[key] for _g in g]) 
                for key in g[0]
            }

        return _parameter_gradient_with_params
    
    def parameter_hessian_diagonal(
        self, 
        model: Optional[StatelessModel] = None,
        approximator: str = _DEFAULT_APPROXIMATOR, 
        *args, **kwargs
    ) -> Callable:
        (
            stateless_model, 
            params, 
            flat_params,
        ) = self.make_stateless_model(model, flat_params=True)
        if approximator in ('bekas', 'exact_diagonal'):
            # these need explicit device placement for hvp
            kwargs['device'] = self.device
        hessian_fn = parameter_hessian_diagonal_unrolled(
            stateless_model,
            approximator=approximator,
            *args, **kwargs,
        )
        packing_fn = partial(vmap, in_axes=(0, None))(self._pack_params_like)

        @jit
        def _parameter_hessian_with_params(x):
            return packing_fn(torch.stack(hessian_fn(flat_params, x)), params)

        return _parameter_hessian_with_params


"""Custom types."""

from typing import TYPE_CHECKING, Any, Iterable, Mapping, Union

from datasets import Dataset, IterableDataset
from pandas import DataFrame
from numpy.typing import ArrayLike

if TYPE_CHECKING:
    from .preprocessing.serializing import Preprocessor
else:
    Preprocessor = Any

DataLike = Union[
    str, 
    DataFrame, 
    Mapping[str, ArrayLike], 
    Dataset, 
    IterableDataset
]
FeatureLike = Union[
    str, 
    Mapping[str, Preprocessor], 
    Mapping[str, Mapping[str, Any]], 
    Iterable[Union[
        str, 
        Mapping[str, Preprocessor], 
        Mapping[str, Mapping[str, Any]]
    ]]
]
StrOrIterableOfStr = Union[str, Iterable[str]]



"""Generating approaximate spectral splits on chemical data using FAISS index."""

from typing import Dict, Iterable, Optional, Mapping, Union

from datasets import Dataset, IterableDataset
import faiss
import numpy as np
from schemist.features import calculate_feature
from tqdm.auto import tqdm, trange

from .bin_packing import pack_bins
from .decorators import process_splits
from .disjoint_set import NumpyDisjointSet
from .utils import annotate_split, dataset_len


def _morgan_fingerprint(
    x: Mapping[str, Iterable],
    structure_column: str, 
    input_representation: str = 'smiles',
    fingerprint_column: str = "fp"
) -> np.ndarray:
    """Compute Morgan fingerprints and pack bits into uint8 array.

    Examples
    ========
    >>> batch = {'smiles': ['CCO', 'CCC']}
    >>> out = _morgan_fingerprint(batch, structure_column='smiles')
    >>> 'fp' in out
    True
    >>> isinstance(out['fp'], np.ndarray)
    True
    >>> out['fp'].shape == (2, 2048 // 8)
    True

    """
    fingerprints, _ = calculate_feature(
        strings=x[structure_column],
        feature_type="fp",
        return_dataframe=False,
        on_bits=False,
    )
    x[fingerprint_column] = np.packbits(fingerprints.astype(bool), axis=-1)
    return x


def _annotate_component(
    x: Mapping[str, str], 
    idx: int,
    disjoint_set: NumpyDisjointSet,
    component_column: str = "faiss_component"
) -> Dict[str, ...]:
    """Tag each entry with its graph component.

    """
    x[component_column] = disjoint_set.find(idx)
    return x


@process_splits
def faiss_split(
    ds: Union[Dataset, IterableDataset],
    structure_column: str,
    input_representation: str = 'smiles',
    fingerprint_column: str = "_faiss_fp_",
    n_neighbors: int = 10,
    batch_size: int = 1024,
    splits: Optional[Mapping[str, float]] = None,
    deterministic: bool = True,
    seed: int = 42,
    cache: str = "./cache",
    gpu: bool = False
):
    """Approximate spectral split using FAISS and union-find.

    For an empty dataset, returns empty splits.

    """
    faiss_index = faiss.IndexBinaryFlat(2048)  # Hamming/Jaccard on 2048-bit vectors
    ds = (
        ds
        .map(
            _morgan_fingerprint,
            fn_kwargs={
                "structure_column": structure_column,
                "input_representation": input_representation,
                "fingerprint_column": fingerprint_column,
            },
            batched=True,
            batch_size=batch_size,
            desc="Annotating fingerprint",
        )
    )
    ds = ds.add_faiss_index(
        column=fingerprint_column, 
        index_name=f"{fingerprint_column}_idx",
        custom_index=faiss_index,
        device=0 if gpu else None,
        dtype=np.uint8,
    )
    # ds.save_faiss_index(
    #     f"{fingerprint_column}_idx", 
    #     index_filename,
    # )   # permanent on-disk index

    N = dataset_len(ds)
    djs = NumpyDisjointSet(N)

    iterator = tqdm(
        ds.iter(batch_size=batch_size),
        desc=f"Finding {n_neighbors} nearest neighbors per molecule",
        total=np.ceil(N / batch_size).astype(int),
    )
    for i, batch in enumerate(iterator):
        start = i * batch_size
        batch_fp = np.asarray(batch[fingerprint_column]).astype(np.uint8)
        scores, indices = ds.search_batch(
            f"{fingerprint_column}_idx", 
            batch_fp, 
            k=n_neighbors,
        )
        for row_of_batch, neighbor_id_list in enumerate(indices):
            row_of_dataset = start + row_of_batch
            # skip neighbor_id_list[0] because it is itself
            for neighbor_id in neighbor_id_list[1:]:
                djs.merge(row_of_dataset, int(neighbor_id))

    # path-compression pass
    for i in trange(N, desc="Compressing search path"):
        djs.parent[i] = djs.find(i)

    # prevents pickling error in .map()
    ds.drop_index(f"{fingerprint_column}_idx") 

    if isinstance(ds, Dataset):
        desc1 = {"desc": "Annotating components"}
        desc2 = {"desc": "Annotating splits"}
    else:
        desc1 = desc2 = {}
    ds = ds.map(
        _annotate_component,
        fn_kwargs={
            "disjoint_set": djs,
            "component_column": "faiss_component",
        },
        with_indices=True, 
        batched=False,
        **desc1
    )

    component_to_split = pack_bins(
        ds,
        group_column="faiss_component",
        splits=splits,
        batch_size=batch_size,
        num_rows=N,
        deterministic=deterministic,
        seed=seed,
    )

    ds = (
        ds
        .map(
            annotate_split,
            fn_kwargs={
                "key_to_split": component_to_split,
                "key_column": "faiss_component",
                "split_column": "split",
            },
            batched=True,
            batch_size=batch_size,
            **desc2
        )
    )

    if isinstance(ds, Dataset):
        desc = {"desc": "Generating split datasets"}
    else:
        desc = {}
    return ds, {
        key: ds.filter(lambda x: x["split"] == key, **desc)
        for key in splits
    }



"""Utilities for dealing with Hugging Face datasets."""

from typing import TYPE_CHECKING, Any, Mapping, Optional, Tuple, Union

import os
from tempfile import TemporaryDirectory

from carabiner import print_err

if TYPE_CHECKING:
    from datasets import (
        Dataset, 
        DatasetDict, 
        IterableDataset
    )
else:
    Dataset, DatasetDict, IterableDataset = Any, Any, Any

import numpy as np
from tqdm.auto import tqdm

from ..checkpoint_utils import save_json

def to_dataset(
    ds: IterableDataset,
    batch_size: int = 1024,
    nrows: Optional[int] = None,
    cache: str = "./cache"
) -> Dataset:
    from datasets import Dataset, concatenate_datasets
    if isinstance(ds, Dataset):
        return ds
    new_ds = None
    total_iter = np.ceil(nrows / batch_size).astype(int) if nrows is not None else None
    for record in tqdm(ds.iter(batch_size=batch_size), total=total_iter, desc="Building dataset"):
        # print(record)
        if new_ds is not None:
            with TemporaryDirectory() as tmpdirname:
                filename = os.path.join(tmpdirname, "add-record.json")
                save_json(record, filename)
                print(record)
                new_ds = concatenate_datasets([new_ds, Dataset.from_json(filename, cache_dir=cache)])
        else:
            with TemporaryDirectory() as tmpdirname:
                filename = os.path.join(tmpdirname, "init-record.json")
                save_json(record, filename)
                print(record)
                new_ds = Dataset.from_json(filename, cache_dir=cache)
    return new_ds


def top_n(
    ds: Union[Dataset, IterableDataset],
    column: str,
    n: int = 100,
    ds_rows: Optional[int] = None
) -> list:
    buffer = set()
    for record in tqdm(ds, total=ds_rows):
        if len(buffer) < n:
            buffer.add(record[column])
        else:
            test_value = record[column]
            for item in buffer:
                if test_value < item:
                    buffer.remove(item)
                    buffer.add(test_value)
                    break
    return list(buffer)


def _stream_and_subsample(
    ds: Dataset,
    min_subsample: int,
    total_rows: int,
    subsample: Optional[int] = None,
    seed: int = 0,
    shuffle_buffer: int = 10_000,
    n_shards: int = 1024
) -> IterableDataset:

    ds_stream = (
        ds
        .to_iterable_dataset(num_shards=n_shards)
        .shuffle(seed=seed, buffer_size=shuffle_buffer)
    )
    
    if subsample is not None:
        if subsample < min_subsample:
            subsample = min_subsample + 1
        if subsample <= total_rows:
            ds_stream = ds_stream.take(subsample)
    else:
        subsample = total_rows

    return subsample, ds_stream


def _get_cutoff(
    ds: Union[Dataset, IterableDataset],
    total_rows: int,
    column: str,
    n_hits: int,
    subsample: Optional[int] = None
) -> Union[float, int]:
    subsample = subsample or total_rows
    subsample = min(subsample, total_rows)
    top_n_values = top_n(
        ds.take(subsample),
        column,
        n=int(n_hits),
        ds_rows=subsample,
    )
    return max(top_n_values)
    
    
def split_3way(
    ds: Union[Dataset, IterableDataset],
    n_train: Optional[int] = None, 
    n_val: int = 0, 
    n_test: Optional[int] = None, 
    filter_column: Optional[str] = None,
    ascending: bool = False,
    sample_for_cutoff: Optional[int] = None,
    hit_frac: float = .01,
    initial_subsample: Optional[int] = None, 
    seed: int = 0,
    shuffle_buffer: int = 10_000,
    n_shards: int = 1024
):  
    from datasets import DatasetDict, concatenate_datasets

    total_rows = ds.num_rows
    n_train = n_train or total_rows
    n_train = n_train if n_train <= total_rows else total_rows
    n_train_and_val = n_train + n_val
    assert (n_train_and_val <= total_rows)
    n_test = n_test or (total_rows - n_train - n_val)

    ds_stream = (
        ds
        .to_iterable_dataset(num_shards=n_shards)
        .shuffle(seed=seed, buffer_size=shuffle_buffer)
    )
    
    if initial_subsample is not None:
        if initial_subsample < n_train_and_val:
            initial_subsample = n_train_and_val + 1
        if initial_subsample <= total_rows:
            ds_stream = ds_stream.take(initial_subsample)
    else:
        initial_subsample = total_rows
        
    if filter_column is not None:
        ds_train_and_val = ds_stream.take(n_train_and_val)
        ds_potential_test = ds_stream.skip(n_train_and_val)#.take(ds_shuffled.num_rows - n_train_and_val)
        ds_potential_test_nrows = initial_subsample - n_train_and_val
        n_hits = int(np.ceil(hit_frac * n_test))
        print_err(f">> Getting hit cutoff from {sample_for_cutoff} / {ds_potential_test_nrows} rows")

        hit_cutoff = _get_cutoff(
            ds=ds_potential_test,
            total_rows=ds_potential_test_nrows,
            column=filter_column,
            n_hits=n_hits,
            subsample=sample_for_cutoff
        )
        print_err(f">>> Hit cutoff is {hit_cutoff:.2f}")
        print_err(f">> Taking {n_hits} rows with {filter_column} < {hit_cutoff:.2f}")
        ds_hits = (
            ds_potential_test
            # .with_format("numpy")
            .filter(
                lambda x: x[filter_column] <= hit_cutoff,
                batched=False,
                # desc=f"Filtering {filter_column} <= {hit_cutoff:.2f}"
            )
            .shuffle(seed=seed, buffer_size=shuffle_buffer)
            .take(n_hits)
        )
        ds_test = (
            ds_potential_test
            # .with_format("numpy")
            .filter(
                lambda x: x[filter_column] > hit_cutoff,
                batched=False,
                # desc=f"Filtering {filter_column} > {hit_cutoff:.2f}"
            )
            .shuffle(seed=seed, buffer_size=shuffle_buffer)
            .take(n_test - n_hits)
        )
        ds_test = (
            concatenate_datasets([ds_hits, ds_test])
            .shuffle(seed=seed, buffer_size=shuffle_buffer)
        )
        ds = {
            "train": to_dataset(ds_train_and_val.take(n_train), ds_rows=n_train),
            "val": to_dataset(ds_train_and_val.skip(n_train), ds_rows=n_val),
            "test": to_dataset(ds_test, ds_rows=n_test),
        }
    else:
        ds = {
            "train": to_dataset(ds_stream.take(n_train), ds_rows=n_train),
            "val": to_dataset(ds_stream.skip(n_train).take(n_val), ds_rows=n_val),
            "test": to_dataset(ds_stream.skip(n_train + n_val).take(n_test), ds_rows=n_test),
        }
    return DatasetDict(ds)


def split_al_pools(
    ds: Dataset, 
    n_initial: int, 
    candidate_filter: Optional[Mapping] = None,
    n_candidates: Optional[int] = None, 
    n_val: Optional[int] = None, 
    batch_size: Optional[int] = None, 
    n_test: Optional[int] = None, 
    n_batches: int = 20,
    filter_column: Optional[str] = None,
    hit_frac: float = .05,
    sample_for_cutoff: Optional[int] = None, 
    initial_subsample: Optional[int] = None, 
    ascending: bool = False,
    seed: int = 0,
    shuffle_buffer: int = 10_000,
    n_shards: int = 1024
) -> Tuple[Tuple[int], DatasetDict]:

    from datasets import DatasetDict, concatenate_datasets
    
    total_rows = ds.num_rows

    n_initial = min(n_initial, total_rows)
    batch_size = batch_size or n_initial
    n_candidates = n_candidates or (n_batches * batch_size)
    assert n_candidates > (n_batches * batch_size), "You need at least as many candidates as will be sampled!"
    n_val = n_val or n_initial
    n_test = n_test or n_candidates

    total_rows_final = n_initial + n_candidates + n_val + n_test

    initial_subsample, ds_stream = _stream_and_subsample(
        ds,
        min_subsample=total_rows_final,
        total_rows=total_rows,
        subsample=initial_subsample,
        shuffle_buffer=shuffle_buffer,
        n_shards=n_shards,
        seed=seed,
    )
    n_noncandidates = total_rows_final - n_candidates

    if candidate_filter is not None:
        ds_potential_candidates = ds_stream
        ds_noncandidates = ds_stream
        for key, val in candidate_filter.items():
            ds_potential_candidates = ds_potential_candidates.filter(
                lambda x: x[key] < val,
            )
            ds_noncandidates = ds_noncandidates.filter(
                lambda x: x[key] >= val,
            )
        ds_noncandidates = ds_noncandidates.take(n_noncandidates)
    else:
        ds_noncandidates = ds_stream.take(n_noncandidates)
        ds_potential_candidates = ds_stream.skip(n_noncandidates)
        
    if filter_column is not None:        
        ds_potential_candidates_nrows = initial_subsample - n_noncandidates
        print_err(f">> Getting hit cutoff from {sample_for_cutoff} / {ds_potential_candidates_nrows} rows")
        n_hits = int(np.ceil(hit_frac * n_test))

        hit_cutoff = _get_cutoff(
            ds=ds_potential_candidates,
            total_rows=ds_potential_candidates_nrows,
            column=filter_column,
            n_hits=n_hits,
            subsample=sample_for_cutoff
        )
        print_err(f">>> Hit cutoff is {hit_cutoff:.2f}")
        print_err(f">> Taking {n_hits} rows with {filter_column} < {hit_cutoff:.2f}")
        ds_hits = (
            ds_potential_candidates
            # .with_format("numpy")
            .filter(
                lambda x: x[filter_column] <= hit_cutoff,
                batched=False,
                # batch_size=10_000,
                # desc=f"Filtering {filter_column} <= {hit_cutoff:.2f}"
            )
            .shuffle(seed=seed, buffer_size=shuffle_buffer)
            .take(n_hits)
        )
        ds_candidates = (
            ds_potential_candidates
            # .with_format("numpy")
            .filter(
                lambda x: x[filter_column] > hit_cutoff,
                batched=False,
                # desc=f"Filtering {filter_column} > {hit_cutoff:.2f}"
            )
            .shuffle(seed=seed, buffer_size=shuffle_buffer)
            .take(int(n_candidates - n_hits))
        )
        ds_candidates = (
            concatenate_datasets([ds_hits, ds_candidates])
            .shuffle(seed=seed, buffer_size=shuffle_buffer)
            .take(n_candidates)
        )#.flatten_indices()
        ds = {
            "initial": to_dataset(ds_noncandidates.take(n_initial), ds_rows=n_initial),
            "candidates": to_dataset(ds_candidates, ds_rows=n_candidates),
            "val": to_dataset(ds_noncandidates.skip(n_initial).take(n_val), ds_rows=n_val),
            "test": to_dataset(ds_noncandidates.skip(n_initial + n_val).take(n_test), ds_rows=n_test),
        }
    else:
        ds = {
            "initial": to_dataset(ds_noncandidates.take(n_initial), ds_rows=n_initial),
            "candidates":  to_dataset(ds_potential_candidates.take(n_candidates), ds_rows=n_candidates),
            "val":  to_dataset(ds_noncandidates.skip(n_initial).take(n_val), ds_rows=n_val),
            "test":  to_dataset(ds_noncandidates.skip(n_initial + n_val).take(n_test), ds_rows=n_test),
        }
    return tuple(range(n_candidates)), DatasetDict(ds)



"""Implementation of Union-Find Tree with numpy."""

import numpy as np

class NumpyDisjointSet:
    """A Union-Find (Disjoint-Set) for elements 0..N-1.
    
    With path-halving find and union-by-size.

    """
    def __init__(self, n: int, memmap: bool = False, cache: str = "parent.npy"):
        self.size = np.ones(n, dtype=np.int32)
        self.memmap = memmap
        if self.memmap:
            self.parent = np.memmap(
                cache,
                dtype=np.int32, 
                mode="w+", 
                shape=(n,),
            )
            self.parent[:] = np.arange(n, dtype=np.int32)
        else:
            self.parent = np.arange(n, dtype=np.int32)

    def find(self, i: int) -> int:
        while self.parent[i] != i:
            self.parent[i] = self.parent[self.parent[i]]
            i = self.parent[i]
        return i

    def merge(self, i: int, j: int) -> None:
        ri, rj = self.find(i), self.find(j)
        if ri == rj:
            return
        if self.size[ri] < self.size[rj]:
            ri, rj = rj, ri
        self.parent[rj] = ri
        self.size[ri] += self.size[rj]

    def __getitem__(self, i: int) -> int:
        return self.find(i)



"""Base mixins for data."""

from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Mapping, Tuple, Optional, Union

from abc import abstractmethod, ABC
from functools import partial
import os

from carabiner import cast, print_err

if TYPE_CHECKING:
    from datasets import Dataset, DatasetDict, IterableDataset
    from pandas import DataFrame
else:
    Dataset, DatasetDict, IterableDataset, DataFrame = Any, Any, Any, Any

import numpy as np
from numpy.typing import ArrayLike
from schemist.converting import convert_string_representation

from ..checkpoint_utils import load_checkpoint_file, save_json
from .preprocessing import Preprocessor
from .typing import DataLike, FeatureLike, StrOrIterableOfStr

_DEFAULT_BATCH_SIZE: int = 128


class DataMixinBase(ABC):

    """Add data-loading capability to models.
    
    """

    _default_cache: str = "cache/duvida/data"
    _default_preprocessing_args: dict = {}
    _in_key: str = 'inputs'
    _out_key: str = 'labels'
    _input_training_data = None
    _input_featurizers = None
    _input_cols = None
    _label_cols = None
    _format: str = 'numpy'
    _format_kwargs: Optional[Mapping[str, Any]] = None
    training_data = None
    training_example = None
    input_shape = None 
    output_shape = None

    def save_data_checkpoint(
        self, 
        checkpoint_dir: str
    ):
        keys = (
            "_in_key",
            "_out_key",
            "_input_cols",
            "_label_cols",
            "_input_featurizers",
            "input_shape",
            "output_shape",
            "_default_cache",
            "_default_preprocessing_args",
        )
        data_config = {
            key: getattr(self, key) for key in keys
        }
        if not os.path.exists(checkpoint_dir):
            os.makedirs(checkpoint_dir)
        if self._input_training_data is not None:
            self._input_training_data.save_to_disk(
                os.path.join(os.path.join(checkpoint_dir, "input-data.hf")),
            )
            if self.training_data is not None:
                (
                    self.training_data
                    .with_format("numpy", dtype="float")
                    .save_to_disk(os.path.join(checkpoint_dir, "training-data.hf")),
                )
        save_json(data_config, os.path.join(checkpoint_dir, "data-config.json"))
        return None

    def load_data_checkpoint(
        self, 
        checkpoint: str,
        cache_dir: Optional[str] = None
    ):
        data_config = load_checkpoint_file(
            checkpoint, 
            filename="data-config.json",
            callback="json",
            none_on_error=False,
            cache_dir=cache_dir,
        )
        for key, val in data_config.items():
            setattr(self, key, val)
        self._input_training_data = load_checkpoint_file(
            checkpoint, 
            filename="input-data.hf",
            callback="hf-dataset",
            none_on_error=True,
            cache_dir=cache_dir,
        )
        training_data = load_checkpoint_file(
            checkpoint, 
            filename="training-data.hf",
            callback="hf-dataset",
            none_on_error=True,
            cache_dir=cache_dir,
        )
        if training_data is not None:
            self.training_data = training_data.with_format(
                self._format, 
                **self._format_kwargs,
            )
            self.training_example = (
                self.training_data
                .take(1)
                .with_format('numpy')
            )
        return self

    @staticmethod
    def _concat_features(
        x: Mapping[str, ArrayLike],
        inputs: StrOrIterableOfStr,
        labels: StrOrIterableOfStr,
        _in_key: str = "inputs",
        _out_key: str = "labels"
    ) -> Dict[str, np.ndarray]:
        
        if isinstance(inputs, str):
            x = {
                _in_key: x[inputs],
                _out_key: [np.asarray(x[col]) for col in cast(labels, to=list)]
            }
            x[_out_key] = np.concatenate([
                col if col.ndim > 1 else col[..., np.newaxis] 
                for col in x[_out_key]
            ], axis=-1)
            return x
        else:
            cols_to_concat = {
                _in_key: cast(inputs, to=list),
                _out_key: cast(labels, to=list),
            }
            x = {
                key: [np.asarray(x[col]) for col in columns]
                for key, columns in cols_to_concat.items()
            }
            return {
                key: np.concatenate([
                    col if col.ndim > 1 else col[..., np.newaxis] 
                    for col in columns
                ], axis=-1)
                for key, columns in x.items()
            }

    @staticmethod
    def _check_is_calculated(
        x: Dataset,  
        preprocessor: Preprocessor
    ) -> Tuple[str, bool]:
        out_column = preprocessor.output_column
        return out_column, out_column in x.column_names

    @staticmethod
    def _featurize(
        x: Mapping[str, ArrayLike],
        featurizers: Iterable[Mapping[str, Any]]
    ) -> Dict[str, np.ndarray]:

        for featurizer in featurizers:
            if isinstance(featurizer, Mapping):
                featurizer = Preprocessor.from_dict(featurizer)
            elif not isinstance(featurizer, Preprocessor):
                raise ValueError(f"Featurizer is neither a dict nor a `Preprocessor`: {featurizer}")
            if featurizer.output_column not in x:
                x = featurizer(x)
            else:
                print_err(f"INFO: {featurizer.output_column} already present, skipping.")
        return x

    @staticmethod
    def _check_column_presence(
        features: StrOrIterableOfStr, 
        labels: StrOrIterableOfStr,
        data: Dataset
    ) -> Iterable[str]:
        columns = cast(features, to=list) + cast(labels, to=list)
        data_cols = data.column_names
        absent_columns = [col for col in columns if col not in data_cols]
        if len(absent_columns) > 0:
            raise ValueError(
                f"""
                Requested columns ({', '.join(columns)}) not present in 
                {type(data)}: {', '.join(absent_columns)}.
                """
            )
        return columns
    
    @staticmethod
    def _load_from_csv(
        filename: str,
        cache: Optional[str] = None
    ) -> Dataset:
        from datasets import Dataset

        if filename.endswith((".csv", ".tsv", ".txt", ".csv.gz", ".tsv.gz", ".txt.gz")):
            read_f = partial(
                Dataset.from_csv,
                cache_dir=cache,
                sep="," if filename.endswith((".csv", ".csv.gz")) else "\t",
            )
        elif filename.endswith(".parquet"):
            read_f = partial(Dataset.from_parquet, cache_dir=cache)
        elif filename.endswith(".hf"):
            read_f = Dataset.load_from_disk
        elif filename.endswith(".arrow"):
            read_f = Dataset.from_file
        else:
            raise IOError(f"Could not infer how to open '{filename}' from its extension.")
        return read_f(filename)

    @classmethod
    def _load_from_dataframe(
        cls,
        dataframe: Union[DataFrame, Mapping[str, ArrayLike]],
        cache: Optional[str] = None
    ) -> Dataset:
        from datasets.fingerprint import Hasher
        from pandas import DataFrame

        if cache is None:
            cache = cls._default_cache
            print_err(f"Defaulting to cache: {cache}")
        if not isinstance(dataframe, DataFrame) and isinstance(dataframe, Mapping):
            dataframe = DataFrame(dataframe)

        hash_name = Hasher.hash(dataframe)
        df_temp_file = os.path.join(cache, "df-load", f"{hash_name}.csv")
        df_temp_dir = os.path.dirname(df_temp_file)
        if not os.path.exists(df_temp_dir):
            os.makedirs(df_temp_dir)

        if not os.path.exists(df_temp_file):
            print_err(f"Caching dataframe at {df_temp_file}")
            dataframe.to_csv(df_temp_file, index=False)

        print_err(f"Reloading dataframe from {df_temp_file}")
        return cls._load_from_csv(
            df_temp_file, 
            cache=cache,
        )

    @staticmethod
    def _resolve_featurizers(
        features: FeatureLike
    ):
        transformer_prefix = "transformer://"
        if isinstance(features, (str, Mapping)):
            features = [features]
        resolved_featurizers = []
        for featurizer in features:
            if isinstance(featurizer, str):
                if featurizer.endswith(".json"):
                    featurizer = Preprocessor.from_file(featurizer)
                elif featurizer.startswith(transformer_prefix):
                    ref = featurizer.split(transformer_prefix)[-1]
                    try:
                        ref, col = ref.split(":")
                    except ValueError:
                        raise ValueError(
                            f"""
                            Transformers models should be provided in the format 
                            {transformer_prefix}<ref>:<input-column>[~agg1,agg2]

                            But got: "{featurizer}"
                            """
                        )

                    try:
                        col, aggs = col.split("~")
                    except ValueError:
                        col, aggs = ref, ["mean"]
                    else:
                        aggs = aggs.split(",")

                    featurizer = Preprocessor(
                        name="hf-bart", 
                        input_column=col,
                        kwargs={
                            "ref": ref,
                            "aggregator": aggs,
                        }
                    )
                elif ":" in featurizer and featurizer.split(":")[-1] in Preprocessor.show():
                    try:
                        name, col = featurizer.split(":")
                    except ValueError:
                        raise ValueError(
                            f"""
                            Column mapped to featurizer name should be in the format
                            <column-name>:<featurizer-name>, with only one colon
                            character.

                            But got "{featurizer}"
                            """
                        )
                    else:
                        featurizer = Preprocessor(name=name, input_column=col)
                else:
                    featurizer = Preprocessor(name="identity", input_column=featurizer)
            elif isinstance(featurizer, Mapping):
                featurizer = Preprocessor.from_dict(featurizer)
            elif isinstance(featurizer, Preprocessor):
                pass
            else:
                raise ValueError(
                    f"""
                    Featurizer must be a column name, HF transformers reference,
                    JSON filename, a dict, or `Preprocessor`, but it was a 
                    {type(featurizer)}: {featurizer}
                    """
                )
            resolved_featurizers.append(featurizer)
        return resolved_featurizers

    @staticmethod
    def _resolve_hf_hub_dataset(
        ref: str, 
        cache: str,
    ) -> Dataset:
        from datasets import concatenate_datasets, load_dataset, DatasetDict

        hf_ref_full = ref.split("hf://")[-1]
        hf_ref = hf_ref_full.split("@")[0] if "@" in ref else hf_ref_full
        if ":" in hf_ref_full:
            ds_config, ds_split = hf_ref_full.split("@")[-1].split(":")[:2]
        else:
            ds_config, ds_split = hf_ref_full.split("@")[-1], None
        
        dataset = load_dataset(path=hf_ref, name=ds_config, split=ds_split, cache_dir=cache)
        if isinstance(dataset, DatasetDict):
            dataset = concatenate_datasets([v for key, v in dataset.items()])
        return dataset

    @classmethod
    def _resolve_data(
        cls, 
        data: DataLike, 
        cache: Optional[str] = None
    ) -> Union[Dataset, IterableDataset]:
        from datasets import Dataset, IterableDataset
        from pandas import DataFrame

        if isinstance(data, (Dataset, IterableDataset)):
            dataset = data
        elif isinstance(data, (DataFrame, Mapping)):
            dataset = cls._load_from_dataframe(
                data, 
                cache=cache,
            )
        elif isinstance(data, str):
            if data.startswith("hf://"):
                dataset = cls._resolve_hf_hub_dataset(
                    data,
                    cache=cache,
                )
            else:
                dataset = cls._load_from_csv(
                    data, 
                    cache=cache,
                )
        else:
            raise ValueError(
                """
                Data must be a string, Dataset, dictionary, or Pandas DataFrame.
                """
            )
        return dataset

    def _ingest_data(
        self, 
        data: DataLike,
        features: Optional[FeatureLike] = None, 
        labels: Optional[StrOrIterableOfStr] = None,
        batch_size: int = _DEFAULT_BATCH_SIZE,
        cache: Optional[str] = None,
        one_column_input: Optional[str] = None,
        _extra_cols_to_keep: Optional[StrOrIterableOfStr] = None,
        **preprocessing_args
    ) -> Tuple[
        List[str], 
        List[str], 
        Dict[str, Dict[str, Union[str, Preprocessor]]], 
        Dataset, 
        Dataset
    ]:

        """Process data to be consistent with training data.
        
        """
        if cache is None:
            cache = self._default_cache
        if features is None:
            features = self._input_featurizers
        if labels is None:
            labels = self._label_cols
        if features is None:
            raise AttributeError(
                """
                You cannot process new data before loading the training data.
                Try running .load_training_data() first.
                """
            )
        elif not isinstance(features, (Mapping, Iterable, str)):
            raise ValueError(
                """
                Features must be a dict, str, or list of str.
                """
            )
        dataset = self._resolve_data(data)
        featurizers = self._resolve_featurizers(features)
        featurizers_dicts = [f.to_dict() for f in featurizers]
        input_columns = sorted(set([
            featurizer.input_column for featurizer in featurizers
        ]))
        if _extra_cols_to_keep is not None:
            input_columns += cast(_extra_cols_to_keep, to=list)
        if len(input_columns) == 0:
            raise AttributeError("No input columns generated for model.")
        labels = cast(labels, to=list)
        preprocessing_args = {
            key: val or self._default_preprocessing_args.get(key)
            for key, val in preprocessing_args.items()
        }
        
        input_dataset = (
            dataset
            .map(
                self.preprocess_data,
                fn_kwargs=preprocessing_args,
                batched=True,
                batch_size=batch_size,
                desc="Preprocessing",
            )
        )
        self._check_column_presence(input_columns, labels, input_dataset)
        input_dataset = (
            input_dataset
            .select_columns(input_columns + labels)
            .map(
                self._featurize,
                fn_kwargs={"featurizers": featurizers_dicts},
                batched=True,
                batch_size=batch_size,
                desc="Featurizing",
            )
        )
        if one_column_input is not None:
            concat_label = one_column_input
        else:
            concat_label = [f.output_column for f in featurizers]
        processed_dataset = (
            input_dataset
            .with_format(None)  # guard against tensors
            .map(
                self._concat_features,
                fn_kwargs={
                    "inputs": concat_label, 
                    "labels": labels,
                    "_in_key": self._in_key,
                    "_out_key": self._out_key,
                },
                batched=True,
                batch_size=batch_size,
                desc="Collating features and labels",
            )
            .select_columns(
                input_columns 
                + labels
                + [self._in_key, self._out_key]
            )
        )

        if self._format_kwargs is None:
            self._format_kwargs = {}

        return (
            input_columns,
            labels,
            featurizers_dicts,
            input_dataset, 
            processed_dataset.with_format(
                self._format, # do not put before concat_features - breaks Arrow with chemprop due to multidim arrays
                **self._format_kwargs,
            )
        )

    def load_training_data(
        self,
        features: FeatureLike, 
        labels: Union[StrOrIterableOfStr, ArrayLike],
        data: DataLike,
        batch_size: int = _DEFAULT_BATCH_SIZE,
        cache: Optional[str] = None,
        **preprocessing_args
    ) -> None:

        """Load dataset used for training.
        
        """

        self._input_cols = cast(features, to=list)
        self._label_cols = cast(labels, to=list)
        (
            self._input_cols, 
            self._label_cols, 
            self._input_featurizers, 
            self._input_training_data, 
            self.training_data,
        ) = self._ingest_data(
            features=features, 
            labels=labels,
            data=data,
            batch_size=batch_size,
            cache=cache,
            **preprocessing_args,
        )
        self.training_example = self.training_data.take(1).with_format('numpy')
        self.input_shape = self.training_example[self._in_key].shape[1:]
        self.output_shape = self.training_example[self._out_key].shape[1:]
        return None

    @staticmethod
    def preprocess_data(data: Mapping[str, ArrayLike]) -> Dict[str, np.ndarray]:
        return data

    @staticmethod
    @abstractmethod
    def make_dataloader(dataset: Iterable, batch_size: int, shuffle: bool):
        pass


class ChemMixinBase(DataMixinBase):

    smiles_column = "clean_smiles"
    common_fp_column = "tanimoto_nn_fp"
    tanimoto_column = "tanimoto_nn"
    
    @staticmethod
    def _featurizer_constructor(
        smiles_column: str,
        use_fp: bool = True,
        use_2d: bool = True,
        extra_featurizers: Optional[FeatureLike] = None,
        _allow_no_features: bool = False
    ) -> Iterable[Preprocessor]:
        featurizer = []
        if all([
            not use_fp, 
            not use_2d,
            extra_featurizers is None or len(extra_featurizers) == 0,
            not _allow_no_features,
        ]):
            print_err("No featurizers defined for fingerprint. Setting `use_fp=True`.")
            use_fp = True
        if use_fp:
            featurizer.append(Preprocessor(
                name="morgan-fingerprint",
                input_column=smiles_column,
            ))
        if use_2d:
            featurizer.append(Preprocessor(
                name="descriptors-2d",
                input_column=smiles_column,
            ))
        if extra_featurizers is not None:
            featurizer += extra_featurizers
        if len(featurizer) == 0 and not _allow_no_features:
            # Should never happen
            raise ValueError("No features defined for fingerprint.")
        else:
            return featurizer

    @staticmethod
    def preprocess_data(
        data: Mapping[str, ArrayLike],
        structure_column: str,
        smiles_column: str,
        input_representation: str = "smiles"
    ) -> Dict[str, np.ndarray]:
        """Generate clean, canonical SMILES.

        Examples
        ========
        >>> d  = {"struc": ["CCC", "CCO"]}
        >>> out = ChemMixinBase.preprocess_data(d, "struc", ChemMixinBase.smiles_column)
        >>> out[ChemMixinBase.smiles_column]
        ['CCC', 'CCO']
        >>> d1 = {"struc": ["CCC"]}  # singleton batch
        >>> out1 = ChemMixinBase.preprocess_data(d1, "struc", ChemMixinBase.smiles_column)
        >>> out1[ChemMixinBase.smiles_column]  # returns list
        ['CCC']
        >>> d2 = {"struc": "CCC"}  # non-batched map
        >>> out2 = ChemMixinBase.preprocess_data(d2, "struc", ChemMixinBase.smiles_column)
        >>> out2[ChemMixinBase.smiles_column]  # still returns list
        ['CCC']
        
        """
        converted = convert_string_representation(
            strings=data[structure_column],
            input_representation=input_representation,
            output_representation="smiles",
        )
        if len(data[structure_column]) > 1 and not isinstance(data[structure_column], str):
            data[smiles_column] = list(converted)
        else:
            data[smiles_column] = [converted]
        return data

    @staticmethod
    def _get_max_sim(
        query: ArrayLike, 
        references: ArrayLike,
        aggregator: Callable[[ArrayLike], float] = np.max
    ):
        a_n_b = np.sum(query[np.newaxis] * references, axis=-1, keepdims=True)
        sum_q = np.sum(query)
        similarities = a_n_b / (sum_q + np.sum(references, axis=-1) - np.sum(a_n_b, axis=-1))[..., np.newaxis]
        return aggregator(similarities)

    @staticmethod
    def _get_nn_tanimoto(
        x: Mapping[str, ArrayLike],
        refs_data: Mapping[str, ArrayLike],
        results_column: str,
        _in_key: str,
        _sim_fn: Callable[[ArrayLike, ArrayLike], float],
    ) -> Dict[str, np.ndarray]:
        query_fps = x[_in_key]
        refs = refs_data[_in_key]
        results = [_sim_fn(q, refs) for q in query_fps]
        results = np.stack(results, axis=0)
        x[results_column] = results
        return x

    def tanimoto_nn(
        self, 
        data: DataLike,
        query_structure_column: Optional[str] = None,
        query_input_representation: Optional[str] = None,
        batch_size: int = _DEFAULT_BATCH_SIZE,
        cache: Optional[str] = None,
        **kwargs
    ) -> Dataset:
        """Get Tanimoto similarity of nearest training set data.
    
        """
        if query_structure_column is None:
            query_structure_column = self._default_preprocessing_args["structure_column"]
        if query_input_representation is None:
            query_input_representation = self._default_preprocessing_args["input_representation"]
        fp_preprocessor = Preprocessor(
            name="morgan-fingerprint",
            input_column=self.__class__.smiles_column,
            **kwargs
        )
        query_dataset = self._resolve_data(data, cache=cache)
        common_map_opts = {
            "batched": True,
            "batch_size": batch_size
        }
        fp_map_opts = {
            "function": self._featurize,
            "fn_kwargs": {
                "featurizers": [fp_preprocessor]
            }
        }
        query_fp_col = fp_preprocessor.output_column
        queries = (
            query_dataset
            .map(
                self.preprocess_data,
                fn_kwargs={
                    "smiles_column": self.__class__.smiles_column,
                    "structure_column": query_structure_column,
                    "input_representation": query_input_representation,
                },
                **common_map_opts,
                desc="Converting to clean SMILES",
            )
            .map(
                **fp_map_opts,
                **common_map_opts,
                desc="Calculating query fingerprints",
            )
            .rename_column(query_fp_col, self.common_fp_column)
            .with_format(
                self._format, 
                **self._format_kwargs,
            )
        )

        query_fp_col, query_is_calculated = self._check_is_calculated(
            queries,
            fp_preprocessor,
        )

        ref_fp_col, ref_is_calculated = self._check_is_calculated(
            self._input_training_data,
            fp_preprocessor,
        )
        if not ref_is_calculated:
            refs = (
                self._input_training_data
                .map(
                    **fp_map_opts,
                    **common_map_opts,
                    desc="Calculating reference fingerprints",
                )
            )
        else:
            refs = self._input_training_data
        refs = (
            refs
            .rename_column(ref_fp_col, self.common_fp_column)
            .select_columns([self.smiles_column, self.common_fp_column])
            .with_format(
                self._format, 
                **self._format_kwargs,
            )
        )
        return queries.map(
            self._get_nn_tanimoto,
            fn_kwargs={
                "refs_data": refs,
                "results_column": self.tanimoto_column,
                "_in_key": self.common_fp_column, 
                "_sim_fn": self._get_max_sim,
            },
            **common_map_opts,
            desc="Calculating Tanimoto similarity to nearest training neighbor",
        ).remove_columns(self.common_fp_column)



# from .hessians import (
#     bekas,
#     exact_diagonal,
#     get_approximators, 
#     rough_finite_difference,
#     squared_jacobian
# )
# from .hvp import hvp
# from .information import (
#     doubtscore,
#     fisher_score,
#     fisher_information_diagonal,
#     parameter_gradient,
#     parameter_hessian_diagonal,
#     information_sensitivity
# )



"""Utilities for dealing with pytorch Lightning logs and outputs."""

from glob import glob
import os

def _get_most_recent_lightning_log(
    save_dir: str,
    filename: str,
    name: str = "lightning_logs",
    version_prefix: str = "version"
) -> str:
    path = os.path.join(save_dir, name)
    logs = sorted(glob(os.path.join(path, f"{version_prefix}_*")))
    max_version = max([int(v.split("_")[-1]) for v in logs])
    max_version = os.path.join(path, f"version_{max_version}", filename)
    if os.path.exists(max_version):
        return max_version
    else:
        raise OSError(f"Could not find most recent log: {max_version}")



"""Automated registry of modelboxes."""

from typing import Callable, Dict, List

from .modelboxes import ModelBoxBase

# Registry to map function names to actual implementations
MODELBOX_REGISTRY: Dict[str, ModelBoxBase] = {}
MODELBOX_REGISTRY_INVERTED: Dict[ModelBoxBase, str] = {}
MODELBOX_NAMES: List[str] = []
DEFAULT_MODELBOX: str = "mlp"


def register_modelbox(name: str) -> Callable:
    """Decorator to register a function in the function registry.
    
    """
    def decorator(c: ModelBoxBase) -> ModelBoxBase:
        c.class_name = name           
        MODELBOX_REGISTRY[name] = c
        MODELBOX_REGISTRY_INVERTED[c] = name
        MODELBOX_NAMES.append(name)
        return c
        
    return decorator



"""Containers for models and their training data."""

from typing import Tuple, Optional
import os

import torch

from ...base.modelboxes import (
    ChempropModelBoxBase,
    FingerprintModelBoxBase, 
    ModelBoxBase, 
    ModelBoxWithVarianceBase
)
from ...base.modelbox_registry import register_modelbox
from ...checkpoint_utils import load_checkpoint_file
from ...stateless.config import config

config.set_backend('torch', precision='float')

from ...stateless.typing import Array, ArrayLike
from ..models import ChempropEnsemble, TorchMLPEnsemble
from .data import ChempropDataMixin, DataMixin, TorchChemMixin
from .information import DoubtMixin, ChempropDoubtMixin
from .training import ModelTrainer


class TorchModelBoxBase(ModelBoxBase, DataMixin, DoubtMixin):

    def save_weights(
        self,
        checkpoint: str
    ) -> None:
        torch.save(
            self.model.state_dict(), 
            os.path.join(checkpoint, "params.pt"),
        )
        return None

    def load_weights(
        self,
        checkpoint: str,
        cache_dir: Optional[str] = None
    ):
        state_dict = load_checkpoint_file(
            checkpoint, 
            filename="params.pt",
            callback="pt",
            none_on_error=False,
            cache_dir=cache_dir,
        )
        self.model.load_state_dict(state_dict)
        return self

    def eval_mode(self) -> None:
        self.model.eval()
        return None

    def train_mode(self) -> None:
        self.model.train()
        return None

    @property
    def shape(self) -> Tuple[Array]:
        return tuple(p.shape for p in self.model.parameters())
    
    @property
    def size(self) -> int:
        return sum(p.numel() for p in self.model.parameters())

    @staticmethod
    def detach_tensor(x: ArrayLike) -> Array:
        return x.detach().cpu().numpy()

    def create_trainer(
        self, 
        epochs: int = 1, 
        **kwargs
    ) -> ModelTrainer:
        trainer_opts = {
            "logger": True,
            # Use `True` if you want to save model checkpoints. The checkpoints will be saved in the `checkpoints` folder.
            "enable_checkpointing": False,
            "enable_progress_bar": True,
            "enable_model_summary": True,
            "accelerator": "auto",
            "devices": "auto",
        }
        trainer_opts.update(kwargs)
        return ModelTrainer(epochs=epochs, **trainer_opts)


@register_modelbox("mlp")
class TorchMLPModelBox(TorchModelBoxBase, ModelBoxWithVarianceBase):

    """ModelBox for pytorch multilayer perceptron ensemble.

    Examples
    ========
    >>> mb = TorchMLPModelBox(ensemble_size=3) 
    >>> mb.input_shape, mb.output_shape = (4,), (1,) # usually set by .load_training_data() 
    >>> mb.model = mb.create_model() 
    >>> mb.model.n_input, mb.model.n_out 
    (4, 1)

    """

    def create_model(self, *args, **kwargs) -> TorchMLPEnsemble:
        self._model_config.update(kwargs)
        return TorchMLPEnsemble(
            n_input=self.input_shape[-1],
            n_out=self.output_shape[-1], 
            *args, **self._model_config,
        )


@register_modelbox("fingerprint")
class TorchFingerprintModelBox(TorchChemMixin, FingerprintModelBoxBase, TorchMLPModelBox):
    pass


@register_modelbox("chemprop")
class ChempropModelBox(ChempropDataMixin, ChempropDoubtMixin, ChempropModelBoxBase, TorchFingerprintModelBox):

    def create_model(self, *args, **kwargs) -> ChempropEnsemble:
        self._model_config.update(kwargs)
        if self.training_example[self._in_key][0]["x_d"] is None:
            self.input_shape = (0,)
        else:
            self.input_shape = self.training_example[self._in_key][0]["x_d"].shape
        return ChempropEnsemble(
            n_input=self.input_shape[-1],
            n_out=self.output_shape[-1], 
            **self._model_config,
        )



"""Mixins for training."""

from lightning import LightningModule, Trainer
from lightning.pytorch.plugins.environments import LightningEnvironment
from torch.utils.data import DataLoader

from ...base.training import ModelTrainerBase
from ...stateless.config import config

config.set_backend('torch', precision='float')

class ModelTrainer(ModelTrainerBase):

    _trainer: Trainer = None

    def create_trainer(self) -> None:
        kwargs = {"plugins": LightningEnvironment()} | self._kwargs
        self._trainer = Trainer(
            max_epochs=self.epochs,
            **kwargs,
        )
        return None

    def train(
        self, 
        model: LightningModule, 
        train_dataloader: DataLoader, 
        val_dataloader: DataLoader
    ) -> None:
        if self._trainer is None:
            self.create_trainer()
        self._trainer.fit(model, train_dataloader, val_dataloader)
        return model



"""Tools to make allowed preprocessing functions."""

from typing import Callable, Dict

# Registry to map function names to actual implementations
FUNCTION_REGISTRY: Dict[str, Callable] = {}


def register_function(name: str):
    """Decorator to register a function in the function registry.
    
    """
    def decorator(f: Callable):             
        FUNCTION_REGISTRY[name] = f
        return f
        
    return decorator



"""Tools for loading and writing reusable package data."""

import os

DUVIDA_CACHE = "DUVIDA_CACHE"
DEFAULT_CACHE = os.path.join("~", ".cache", "duvida", "data")

def _get_data_path(
    filename: str, 
    env_key: str = DUVIDA_CACHE,
    default: str = DEFAULT_CACHE
) -> str:
    """Returns the path to a writable version of a package data file.
    
    Copies it from the package resources if not present.

    """
    cache_dir = os.environ.get(
        env_key, 
        os.path.expanduser(default),
    )
    os.makedirs(cache_dir, exist_ok=True)
    return cache_dir, os.path.join(cache_dir, filename)



"""Functions for collapsing dimensions."""

from typing import Callable, Union
from functools import partial

from numpy import asarray, nanmean, nanmedian, nansum, nanvar, ndarray, take
from numpy.linalg import norm
from numpy.typing import ArrayLike

AggFunction = Callable[[ArrayLike], ndarray]

AGGREGATORS = {
    "mean": nanmean,
    "median": nanmedian,
    "sum": nansum,
    "var": nanvar,
    "rms": partial(norm, ord=2, keepdims=False),
}


def slice_at_axis(
    a: ArrayLike,
    i: int, 
    axis: int = -1
):
    # n_dims = len(a.shape)
    # if axis < 0:
    #     axis = n_dims + axis
    # slicer = tuple([] if j == axis else [i] for j in range(n_dims))
    return take(a, asarray(i).astype(int), axis=axis)
    

def get_aggregator(
    aggregator: Union[str, AggFunction],
    axis: int = -1,
    **kwargs
) -> AggFunction:
    
    """Aggregate an array along an axis.

    Examples
    ========
    >>> import numpy as np
    >>> a = np.array([[3., 4.]])
    >>> rms = get_aggregator("rms", axis=-1)
    >>> norm2 = get_aggregator("2-norm", axis=-1)
    >>> rms(a)
    array([5.])
    >>> rms(a) == norm2(a)
    array([ True])
    >>> slicer_0 = get_aggregator(0, axis=-1) 
    >>> slicer_0(np.array([[1., 2.], [3., 4.]])) 
    array([1., 3.])

    """

    if isinstance(aggregator, int):
        return partial(slice_at_axis, i=aggregator, axis=axis)
    elif isinstance(aggregator, str):
        if aggregator.endswith("-norm") and aggregator.split("-")[0].isdigit():
            return partial(norm, ord=int(aggregator.split("-")[0]), axis=axis)
        elif aggregator in AGGREGATORS:
            return partial(AGGREGATORS[aggregator], axis=axis, **kwargs)
        else:
            raise ValueError(f"Invalid aggregator name '{aggregator}'!")
    elif isinstance(aggregator, Callable):
        return aggregator
    else:
        raise ValueError(f"Invalid aggregator type '{type(aggregator)}'! Must be a str or callable.")



"""Mixins to enable Lightning features."""

from typing import Mapping, Optional

from lightning import LightningModule
from torch.optim import Optimizer
from torch.optim.lr_scheduler import ReduceLROnPlateau 

from ...functions import mse_loss
from ....stateless.typing import Array, ArrayLike


class LightningMixin(LightningModule):

    # optimizer: Optimizer = None
    # learning_rate: float = .01
    model_attr: str = None
    # reduce_lr_on_plateau: bool = False
    # reduce_lr_patience: int = 10

    def _init_lightning(
        self, 
        model_attr: str,
        optimizer: Optional[Optimizer] = None, 
        learning_rate: float = .01,
        reduce_lr_on_plateau: bool = False,
        reduce_lr_patience: int = 10
    ) -> None:
        if issubclass(optimizer, Optimizer):
            self.optimizer = optimizer
        else:
            raise ValueError(
                "Optimizer must by a PyTorch Optimizer, but "
                f"'{optimizer}' inherits from \'{' / '.join(optimizer.__mro__)}\'.")
        self.learning_rate = learning_rate
        if model_attr in dir(self):
            self.model_attr = model_attr
        else:
            raise KeyError(f"The model attribute '{model_attr}' does not exist!")
        self.reduce_lr_on_plateau = reduce_lr_on_plateau
        self.reduce_lr_patience = reduce_lr_patience
        return None

    def get_loss(
        self, 
        batch: Mapping[str, ArrayLike]
    ) -> Array:
        inputs, outputs = batch['inputs'], batch['labels']
        predicted = self(inputs)
        return mse_loss(predicted, outputs)

    def training_step(
        self, 
        batch: Mapping[str, ArrayLike], 
        batch_idx: int
    ) -> Array:
        loss = self.get_loss(batch)
        self.log('loss', loss, on_step=False, on_epoch=True, prog_bar=True, batch_size=batch['labels'].shape[-1])
        return loss

    def validation_step(
        self, 
        batch: Mapping[str, ArrayLike], 
        batch_idx: int
    ) -> Array:
        loss = self.get_loss(batch)
        self.log('val_loss', loss, on_step=False, on_epoch=True, prog_bar=True, batch_size=batch['labels'].shape[-1])
        return loss

    def on_validation_epoch_end(self):
        # Log the learning rate.
        if self.reduce_lr_on_plateau:
            lr = self.trainer.lr_scheduler_configs[0].scheduler.get_last_lr()[0]
            self.log('learning_rate', lr, on_step=False, on_epoch=True, prog_bar=True)

    def configure_optimizers(self):
        optimizer = self.optimizer(
            getattr(self, self.model_attr).parameters(), 
            lr=self.learning_rate,
        )
        if self.reduce_lr_on_plateau:
            return {
                "optimizer": optimizer,
                "lr_scheduler": ReduceLROnPlateau(
                    optimizer,
                    patience=self.reduce_lr_patience,
                ),
                "monitor": "val_loss",
            }
        else:
            return optimizer



"""Base mixins for calculating uncertainty and information metrics."""

from typing import Callable, Dict, Iterable, List, Mapping, Optional, Union
from abc import abstractmethod, ABC
from functools import partial, reduce

from carabiner import cast
from datasets import Dataset
from numpy import concatenate, ndarray
from numpy.typing import ArrayLike

from .aggregators import get_aggregator, AggFunction


class DoubtMixinBase(ABC):

    optimizer = None

    @staticmethod
    def _pack_params_like(
        p: ArrayLike, 
        d: Mapping[str, ArrayLike]
    ) -> Dict[str, ndarray]:
        index = 0
        output = {}
        for name, param in d.items():
            end_index = index + param.numel()
            output[name] = p[index:end_index].reshape(*param.shape)
            index = end_index
        return output

    @staticmethod
    @abstractmethod
    def _get_shape(x: ArrayLike):
        pass

    def _add_size(
        self, 
        x: ArrayLike, 
        y: Mapping[str, ArrayLike], 
        axis: int = 0
    ) -> Dict[str, ndarray]:
        return {
            key: self._get_shape(_x)[axis] + self._get_shape(y[key])[axis] 
            for key, _x in y.items()
        }

    @staticmethod
    def _add_values(
        x: ArrayLike, 
        y: Mapping[str, ArrayLike]
    ) -> Dict[str, ndarray]:
        return {key: x[key] + _y for key, _y in y.items()}

    @staticmethod
    def _reduce_dataset(
        x: Dataset, 
        reducer: Callable,
        columns: Optional[Union[str, Iterable[str]]] = None,
        batch_size: int = 1000
    ) -> Dataset:
        if columns is not None:
            columns = [
                col for col in cast(columns, to=list) 
                if col in x.column_names
            ]
        return {col: reducer(x[col]) for col in columns}

    def _mean_of_dictionaries(
        self,
        x: Iterable[Mapping[str, ArrayLike]], 
        axis: int = 0
    ) -> Union[None, Dict[str, ndarray]]:
        n = len(x)
        if n > 0:
            total_size = reduce(self._add_size, x)
            total_value = reduce(self._add_values, x)
            return {key: (val / total_size[key]) for key, val in total_value.items()}
        else:
            return None

    @staticmethod
    @abstractmethod
    def release_memory() -> None:
        pass

    @abstractmethod
    def parameter_gradient(self, *args, **kwargs) -> Callable:
        pass

    @abstractmethod
    def fisher_score(self, *args, **kwargs) -> Callable:
        pass
    
    @abstractmethod
    def parameter_hessian_diagonal(self, *args, **kwargs) -> Callable:
        pass
    
    @abstractmethod
    def fisher_information_diagonal(self, *args, **kwargs) -> Callable:
        pass
    
    @staticmethod
    @abstractmethod
    def doubtscore_core(fisher_score, parameter_gradient):
        pass
    
    @staticmethod
    @abstractmethod
    def information_sensitivity_core(
        fisher_score, 
        fisher_information_diagonal, 
        parameter_gradient, 
        parameter_hessian_diagonal
    ):
        pass
    
    @staticmethod
    def _process_information(
        data: Mapping[str, ArrayLike],
        grad_fns: Iterable[Callable],
        _in_key: str = "inputs",
        _out_key: str = "labels"
    ) -> Dict[str, List[Dict]]:
        grads_hessians = dict(zip(
            ('fisher_score', 'fisher_information_diagonal'),
            [[]] * len(grad_fns),
        ))
        for key, f in zip(grads_hessians, grad_fns):
            f_result = f(x_true=data[_in_key], y_true=data[_out_key])
            grads_hessians[key].append(f_result)
        return grads_hessians

    def _information_scores(
        self, 
        model: Optional[Callable] = None,
        dataset: Optional[Dataset] = None,
        hessian: bool = False, 
        batch_size: int = 16,
        **kwargs
    ) -> Dataset:
        if dataset is None:  
            dataset = self._check_training_data()
        fn = (self.fisher_score,)
        if hessian:
            fn += (partial(
                self.fisher_information_diagonal, 
                **kwargs,
            ),)
        
        grad_fns = tuple(f(model) for f in fn)
        dataset = (
            dataset
            .map(
                self._process_information, 
                fn_kwargs={
                    "grad_fns": grad_fns,
                    "_in_key": self._in_key,
                    "_out_key": self._out_key,
                },
                batch_size=batch_size,
                batched=True,
                remove_columns=dataset.column_names,
                desc=f"Calculating Fisher score{' and information' if hessian else ''}",
            )
        )
        columns = [
            col for col in ['fisher_score', 'fisher_information_diagonal'] 
            if col in dataset.column_names
        ]
        return (
            self._reduce_dataset(dataset, self._mean_of_dictionaries, col)[col]
            for col in columns
        )

    def _doubtscore(
        self, 
        data: Mapping[str, ArrayLike], 
        param_grad_fn: Callable, 
        param_hessian_fn: Callable,
        fisher_score: Mapping[str, ArrayLike], 
        loss_hessian: Optional[Mapping[str, ArrayLike]] = None,
        aggregator: Union[str, AggFunction] = 'rms'
    ) -> Dict[str, ndarray]:
        
        aggregator = get_aggregator(aggregator, axis=-1)
        param_gradient = param_grad_fn(data[self._in_key])
        doubtscore = aggregator(
            concatenate([
                self.doubtscore_core(fisher_score[name], param_gradient[name]) 
                for name in fisher_score
            ], axis=-1), 
        )
        data["score"] = doubtscore
        return data

    def _information_sensitivity(
        self, 
        data: Mapping[str, ArrayLike], 
        param_grad_fn: Callable, 
        param_hessian_fn: Callable,
        fisher_score: Mapping[str, ArrayLike], 
        loss_hessian: Mapping[str, ArrayLike],
        optimality_approximation: bool = False,
        aggregator: Union[str, AggFunction] = 'rms'
    ) -> Dict[str, ndarray]:

        aggregator = get_aggregator(aggregator, axis=-1)
        if optimality_approximation:
            param_gradient, param_hessian = (
                param_grad_fn(data[self._in_key]), 
                {name: None for name in fisher_score}
            )
        else:
            param_gradient, param_hessian = (
                fn(data[self._in_key]) for fn in (param_grad_fn, param_hessian_fn)
            )
        infosens = aggregator(
            concatenate([
                self.information_sensitivity_core(
                    fisher_score[name], 
                    loss_hessian[name], 
                    param_gradient[name], 
                    param_hessian[name],
                    optimality_approximation=optimality_approximation,
                ) for name in fisher_score
            ], axis=-1), 
        )
        data["score"] = infosens
        return data

    def _get_info_score(
        self, 
        score_type: str,
        dataset: Dataset,
        candidates: Dataset,
        batch_size: int = 16,
        model: Optional[Callable] = None,
        optimality_approximation: bool = False,
        **kwargs,
    ) -> Dataset:

        if score_type == "information sensitivity":
            inner_fn = partial(
                self._information_sensitivity, 
                optimality_approximation=optimality_approximation,
            )
        elif score_type == "doubtscore":
            inner_fn = self._doubtscore
        else:
            raise NotImplementedError(f"Score '{score_type}' not yet implemented.")
            
        use_hessian = (score_type == "information sensitivity")
        use_param_hessian = (use_hessian and not optimality_approximation)
        fisher_score_and_info = self._information_scores(
            model=model,
            hessian=use_hessian,
            batch_size=batch_size, 
            **kwargs,
        )
        if use_hessian:
            fisher_score, fisher_info_diag = fisher_score_and_info
        else:
            fisher_score, fisher_info_diag = next(fisher_score_and_info), None

        if use_param_hessian:
            param_hessian_fn = self.parameter_hessian_diagonal(
                model=model,
                **kwargs,
            )
        else:
            param_hessian_fn = None            

        score = candidates.map(
            inner_fn,
            fn_kwargs={
                "param_grad_fn": self.parameter_gradient(model=model), 
                "param_hessian_fn": param_hessian_fn,
                "fisher_score": fisher_score,
                "loss_hessian": fisher_info_diag,
            },
            batched=True, 
            batch_size=batch_size,
            desc=f"Calculating parameter gradients{' and Hessians' if use_hessian and not optimality_approximation else ''}",
        )#.select_columns(['score'])
        return score.rename_column('score', score_type)

    def _check_training_data(self):
        if hasattr(self, "training_data"):
            if getattr(self, "training_data") is None:
                raise AttributeError("Training data is not provided. Run .load_training_data() first!")
            else:
                return self.training_data
        else:
            raise ValueError("Training data is not provided!")

    def _info_score_entrypoint(
        self, 
        score_type: str,
        candidates: Union[ArrayLike, Dataset],
        features: Optional[str] = None,
        labels: Optional[str] = None,
        training_dataset: Optional[Dataset] = None,
        batch_size: int = 16,
        preprocessing_args: Optional[Mapping] = None,
        cache: Optional[str] = None,
        **info_score_kwargs
    ) -> Dataset:
        self.release_memory()
        if preprocessing_args is None:
            preprocessing_args = {}
        if training_dataset is None:  
            dataset = self._check_training_data()

        if hasattr(self, "_prepare_data"):
            candidates = self._prepare_data(
                data=candidates,
                features=features,
                labels=labels,
                batch_size=batch_size,
                cache=cache,
                **preprocessing_args
            )
            
        return self._get_info_score(
            score_type=score_type,
            candidates=candidates,
            dataset=dataset, 
            batch_size=batch_size,
            **info_score_kwargs,
        )

    def doubtscore(
        self, 
        features: Optional[str] = None,
        preprocessing_args: Optional[Mapping] = None,
        **info_score_kwargs
    ) -> Dataset:

        """Calculate doubtscore.
        
        """

        return self._info_score_entrypoint(
            score_type="doubtscore",
            features=features,
            preprocessing_args=preprocessing_args,
            **info_score_kwargs
        )

    def information_sensitivity(
        self, 
        features: Optional[str] = None,
        preprocessing_args: Optional[Mapping] = None,
        **info_score_kwargs
    ) -> Dataset:

        """Calculate information sensitivity.
        
        """

        return self._info_score_entrypoint(
            score_type="information sensitivity",
            features=features,
            preprocessing_args=preprocessing_args,
            **info_score_kwargs
        )



"""Chemprop message-passing neural network."""

from abc import ABC, abstractmethod

from carabiner import print_err
from chemprop.models import MPNN
from chemprop.nn import (
    BondMessagePassing, 
    EvidentialFFN, 
    NormAggregation, 
    RegressionFFN
)
import torch
from torch.nn import Module
from torch.optim import Adam, Optimizer

from .data import (
    DuvidaTrainingBatch, 
    _collate_training_batch_for_forward
)
from ..utils.ensemble import TorchEnsembleMixin
from ..utils.lt import LightningMixin


class ChempropBase(Module, ABC):

    def __init__(
        self, 
        n_input: int, 
        n_hidden: int = 1,
        n_units: int = 16, 
        mp_units: int = 300,
        mp_hidden: int = 3,
        mp_activation: str = "elu",
        learning_rate: float = .01,
        dropout: float = 0., 
        activation: str = "ELU",  # Smooth activation to prevent gradient collapse
        n_out: int = 1, 
        batch_norm: bool = False,
        evidential: bool = False,
        warmup_epochs: int = 2,
        *args, **kwargs
    ):
        super().__init__()
        self._extra_args = args
        self._extra_kwargs = kwargs
        self.n_input = n_input
        self.n_hidden = n_hidden
        self.n_units = n_units
        self.mp_units = mp_units
        self.mp_hidden = mp_hidden
        self.mp_activation = mp_activation
        self.learning_rate = learning_rate
        self.dropout = dropout
        self.activation = activation
        self.n_out = n_out
        self.batch_norm = batch_norm
        self.evidential = evidential
        self.warmup_epochs = warmup_epochs
        self.n_mp_units: int = None

        if self.batch_norm:
            # TODO: fix this issue.
            print_err("Warning: using batch norm, which will not allow calculation of doubtscore or information sensitivity.")

    def build_model(self) -> Module:
        message_passing_layer = BondMessagePassing(
            d_h=self.mp_units,        
            depth=self.mp_hidden,
            activation=self.mp_activation, 
        )
        self.n_mp_units = message_passing_layer.output_dim
        aggregation_layer = NormAggregation()
        predictor_kwargs = {
            "n_tasks": self.n_out,
            "input_dim": self.n_input + self.n_mp_units,
            "hidden_dim": self.n_units,
            "n_layers": self.n_hidden,
            "dropout": self.dropout,
            "activation": self.activation,
        }
        predictor_layer = (
            EvidentialFFN(**predictor_kwargs) if self.evidential 
            else RegressionFFN(**predictor_kwargs)
        )
        return MPNN(
            message_passing_layer, 
            aggregation_layer, 
            predictor_layer, 
            warmup_epochs=self.warmup_epochs,
            batch_norm=self.batch_norm,
            init_lr=self.learning_rate, 
            max_lr=self.learning_rate * 10., 
            final_lr=self.learning_rate,
        )

    @abstractmethod
    def forward(self, x: DuvidaTrainingBatch) -> torch.Tensor:
        pass


class ChempropEnsemble(TorchEnsembleMixin, ChempropBase, LightningMixin):

    def __init__(
        self, 
        ensemble_size: int = 1,
        learning_rate: float = .01,
        optimizer: Optimizer = Adam,
        reduce_lr_on_plateau: bool = True,
        reduce_lr_patience: int = 10,
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.save_hyperparameters()
        self._init_ensemble(ensemble_size)
        self._init_lightning(
            optimizer=optimizer, 
            learning_rate=learning_rate, 
            model_attr='_model_ensemble',
            reduce_lr_on_plateau=reduce_lr_on_plateau,
            reduce_lr_patience=reduce_lr_patience,
        )

    def create_module(self):
        return self.build_model()

    def forward(self, x: DuvidaTrainingBatch) -> torch.Tensor:
        if not isinstance(x, DuvidaTrainingBatch):
            for p in self.parameters():
                device = p.device
                break
            x = _collate_training_batch_for_forward(x, device=device)
        x = (x.bmg, x.V_d, x.X_d)
        # Stack outputs to shape [batch, n_out, ensemble_size]
        out = torch.stack([
            module(*x) for _, module in self.model_ensemble.items()
        ], dim=-1)
        # If n_out == 1 (scalar), squeeze the middle axis to get [batch, ensemble_size]
        if out.size(-2) == 1:
            out = out.squeeze(-2)
        return out



"""Building and managing model ensembles."""

from typing import Iterable, Union

from abc import ABC, abstractmethod

from torch import stack
from torch.nn import Module, ModuleDict

from ....stateless.typing import Array, ArrayLike


class TorchEnsembleMixin(ABC):

    _module_prefix: str = "ensemble_module"

    def _init_ensemble(self, ensemble_size: int) -> None:
        self.ensemble_size = ensemble_size
        self.model_to_use = 'all'
        self.model_keys = None
        self._model_ensemble = self.create_ensemble()
        self.set_model(self.model_to_use)
        return None

    @abstractmethod
    def create_module(self) -> Module:
        pass
        
    def create_ensemble(self) -> ModuleDict:
        return ModuleDict({f"{self._module_prefix}:{i}": self.create_module() for i in range(self.ensemble_size)})

    def set_model(self, n: Union[str, int] = 'all') -> None:
        if n == 'all':
            ensemble = self._model_ensemble
        elif isinstance(n, int) and n < len(self._model_ensemble):
            key = f"{self._module_prefix}:{n}"
            ensemble = ModuleDict({key: self._model_ensemble[key]})
        elif (
            isinstance(n, Iterable) 
            and all(isinstance(_n, int) for _n in n) 
            and all(_n < len(self._model_ensemble) for _n in n)
        ):
            keys = {f"{self._module_prefix}:{i}" for i in n}
            ensemble = ModuleDict({key: self._model_ensemble[key] for key in keys})
        else:
            raise ValueError(f"Cannot set model with n={n} from ensemble with length{(len(self._model_ensemble))}")
        self.model_to_use = n
        self.model_ensemble = ensemble
        self.model_keys = tuple(self.model_ensemble.keys())
        return None

    def forward(self, x: ArrayLike) -> Array:
        # Stack outputs to shape [batch, n_out, ensemble_size]
        out = stack([
            module(x) for _, module in self.model_ensemble.items()
        ], dim=-1)
        # If n_out == 1 (scalar), squeeze the middle axis to get [batch, ensemble_size]
        if out.size(-2) == 1:
            out = out.squeeze(-2)
        return out



"""Preprocessing functions from deep embeddings."""

from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Mapping, Tuple
from functools import partial

from carabiner import cast, print_err
import numpy as np
from numpy.typing import ArrayLike
import torch

if TYPE_CHECKING:
    from transformers import PreTrainedTokenizerBase, PreTrainedModel
else:
    PreTrainedTokenizerBase, PreTrainedModel = Any, Any

from .registry import register_function
from ..typing import StrOrIterableOfStr

_DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def _index_into(x: ArrayLike, i: int):
    return x[:,i,:]


def _get_value_from_tuple(f: Callable, **kwargs):
    partial_f = partial(f, **kwargs)

    def _f(*args, **fkwargs):
        return partial_f(*args, **fkwargs).values

    return _f


EMBEDDING_AGGREGATORS = {
    "start": partial(_index_into, i=0),
    "end": partial(_index_into, i=-1),
    "sum": partial(torch.sum, dim=-2),
    "mean": partial(torch.mean, dim=-2),
    "median": _get_value_from_tuple(torch.median, dim=-2),
    "max": _get_value_from_tuple(torch.max, dim=-2),
}
DEFAULT_AGGREGATOR = EMBEDDING_AGGREGATORS["mean"]


def _aggregated_embedding(
    x: torch.Tensor, 
    aggregators: Iterable[Callable] = (DEFAULT_AGGREGATOR, )
):
    aggregated = [
        a(x).detach().cpu().numpy() for a in aggregators
    ]

    return np.concatenate(aggregated, axis=-1)


def _resolve_bart_model(ref: str) -> Tuple[PreTrainedTokenizerBase, PreTrainedModel]:
    from transformers import AutoModelForSeq2SeqLM, AutoTokenizer
    model = AutoModelForSeq2SeqLM.from_pretrained(ref)
    return AutoTokenizer.from_pretrained(ref), torch.compile(model, fullgraph=True).to(_DEVICE)


def _tokenize_for_embedding(
    x: Iterable[str],
    tokenizer: PreTrainedTokenizerBase
) -> Dict[str, np.ndarray]:

    inputs = [
        _x if _x is not None 
        else '<unk>' 
        for _x in cast(x, to=list)
    ]
    tokenizer_args = {
        "return_tensors": "pt",
        "padding": True,
    }
    try:
        tokenized = tokenizer(inputs, **tokenizer_args)
    except (ValueError, TypeError) as e:
        print_err(e)
        print_err(f"Tokenizing this batch failed: {inputs}")
        tokenized = tokenizer(
            ['<unk>'] * len(inputs), 
            **tokenizer_args,
        )

    return {
        key: tokenized[key].to(_DEVICE) 
        for key in ['input_ids', 'attention_mask']
    }


@register_function("hf-bart")
def HfBART(
    ref: str,
    aggregator: StrOrIterableOfStr = "mean"
) -> np.ndarray:

    aggregators = [a.casefold() for a in cast(aggregator, to=list)]
    incorrect_agg = [
        a for a in aggregators if a not in EMBEDDING_AGGREGATORS
    ]
    if len(incorrect_agg) > 0:
        raise ValueError(f"Requested aggregators do not exist: {incorrect_agg}")
    else:
        aggregators = [EMBEDDING_AGGREGATORS[a] for a in aggregators]
    tokenizer, model = _resolve_bart_model(ref)

    def _hf_bart(
        data: Mapping[str, Iterable],
        input_column: str
    ) -> np.ndarray:
        tokenized_inputs = _tokenize_for_embedding(
            data[input_column],
            tokenizer=tokenizer,
        )
        model.eval()
        with torch.no_grad():
            outputs = model(
                **tokenized_inputs, 
                decoder_input_ids=tokenized_inputs['input_ids'],
                output_hidden_states=True,
                return_dict=True,
            )
        enc_last = _aggregated_embedding(
            outputs.encoder_hidden_states[-1],
            aggregators=aggregators,
        )
        dec_last = _aggregated_embedding(
            outputs.decoder_hidden_states[-1],
            aggregators=aggregators,
        )
        return np.concatenate([enc_last, dec_last], axis=-1)

    return _hf_bart



""""Utilities for loading and saving checkpoints."""

from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Union
from tempfile import TemporaryDirectory
import os
import json

from carabiner import print_err

if TYPE_CHECKING:
    from datasets import Dataset, IterableDataset
else:
    Dataset, IterableDataset = Any, Any


def _load_json(checkpoint: str, filename: str) -> Dict[str, Any]:
    with open(os.path.join(checkpoint, filename), "r") as f:
            obj = json.load(f)
    return obj
   

def save_json(obj, filename: str) -> None:
    with open(filename, "w") as f:
        json.dump(obj, f, sort_keys=True, indent=4)
    return None


def _load_hf_dataset(checkpoint, filename) -> Union[Dataset, IterableDataset]:
    from datasets import load_from_disk
    return load_from_disk(os.path.join(checkpoint, filename))


def _load_torch_weights(checkpoint, filename):
    import torch
    return torch.load(
        os.path.join(checkpoint, filename),
        weights_only=True,
    )


FILE_LOADING_CALLBACKS = {
    "json": _load_json,
    "hf-dataset": _load_hf_dataset, 
    "pt": _load_torch_weights,
}


def load_checkpoint_file(
    checkpoint: str,
    filename: str,
    callback: Union[str, Callable] = "json",
    none_on_error: bool = False,
    cache_dir: Optional[str] = None,
    *args, **kwargs
) -> Union[Any, None]:
    from huggingface_hub import snapshot_download

    obj = None
    if isinstance(callback, str):
        try:
            callback = FILE_LOADING_CALLBACKS[callback.casefold()]
        except KeyError:
            raise ValueError(
                """
                File loading callback must be callable or name.
                """
            )
    if os.path.exists(checkpoint):
        obj = callback(checkpoint, filename)
    elif checkpoint.startswith("hf://"):
        checkpoint = checkpoint.split("hf://")[-1]
        if filename.endswith(".hf"):
            filename_pattern = [filename + '/*.arrow', filename + '/*.json']
        else:
            filename_pattern = filename
        with TemporaryDirectory() as tmpdirname:
            try:
                print_err(f"Looking up: {checkpoint} :: {filename}")
                snapshot_download(
                    repo_id=checkpoint,
                    allow_patterns=filename_pattern,
                    local_dir=tmpdirname,
                    cache_dir=cache_dir,
                    *args, **kwargs
                )
            except Exception as e:
                print_err(e)
                if none_on_error:
                    return None
                else:
                    raise e
            else:
                obj = callback(tmpdirname, filename)
    if obj is not None:
        return obj
    else:
        raise AttributeError(
            f"Could not load anything from {checkpoint=}, {filename=} with {callback=}."
        )



"""Data structures for using chemprop."""

from typing import Iterable, Mapping, Optional, Union
from dataclasses import dataclass, field, InitVar

from chemprop.data import Datum, MolGraph, TrainingBatch
import numpy as np
import torch


@dataclass(repr=False, eq=False, slots=True)
class DuvidaBatchMolGraph:
    """A :class:`BatchMolGraph` represents a batch of individual :class:`MolGraph`s.

    It has all the attributes of a ``MolGraph`` with the addition of the ``batch`` attribute. This
    class is intended for use with data loading, so it uses :obj:`~torch.Tensor`s to store data
    """

    mgs: InitVar[Iterable[MolGraph]]
    """A list of individual :class:`MolGraph`s to be batched together"""
    V: torch.Tensor = field(init=False)
    """the atom feature matrix"""
    E: torch.Tensor = field(init=False)
    """the bond feature matrix"""
    edge_index: torch.Tensor = field(init=False)
    """an tensor of shape ``2 x E`` containing the edges of the graph in COO format"""
    rev_edge_index: torch.Tensor = field(init=False)
    """A tensor of shape ``E`` that maps from an edge index to the index of the source of the
    reverse edge in the ``edge_index`` attribute."""
    batch: torch.Tensor = field(init=False)
    """the index of the parent :class:`MolGraph` in the batched graph"""

    __size: int = field(init=False)

    def __post_init__(self, mgs: Iterable[MolGraph]):
        self.__size = len(mgs)

        Vs = []
        Es = []
        edge_indexes = []
        rev_edge_indexes = []
        batch_indexes = []

        num_nodes = 0
        num_edges = 0
        for i, mg in enumerate(mgs):
            Vs.append(mg.V)
            Es.append(mg.E)
            edge_indexes.append(mg.edge_index + num_nodes)
            rev_edge_indexes.append(mg.rev_edge_index + num_edges)
            # batch_indexes.append(torch.ones((len(mg.V),)) * i)
            batch_indexes.append([i] * len(mg.V))

            num_nodes += mg.V.shape[0]
            num_edges += mg.edge_index.shape[1]

        self.V = torch.concat(Vs)
        try:
            self.E = torch.concat(Es)
        except TypeError:
            Es = [
                E for E in Es if not isinstance(E, list)
            ]
            if len(Es) > 0:
                self.E = torch.concat(Es)
            else:
                self.E = torch.empty((0,))
        self.edge_index = torch.hstack(edge_indexes).long()
        self.rev_edge_index = torch.concat(rev_edge_indexes).long()
        # self.batch = torch.concat(batch_indexes).long()
        self.batch = torch.tensor(np.concatenate(batch_indexes)).long()

    def __len__(self) -> int:
        """the number of individual :class:`MolGraph`s in this batch"""
        return self.__size

    def to(self, device: Union[str, torch.device]):
        self.V = self.V.to(device)
        self.E = self.E.to(device)
        self.edge_index = self.edge_index.to(device)
        self.rev_edge_index = self.rev_edge_index.to(device)
        self.batch = self.batch.to(device)


class DuvidaTrainingBatch(TrainingBatch):

    def to(self, device: Union[str, torch.device]) -> TrainingBatch:
        for p in self:
            if isinstance(p, (torch.Tensor, DuvidaBatchMolGraph)):
                p.to(device)
        return self


def collate_batch(
    batch: Iterable[Datum], 
    device: Optional[Union[str, torch.device]] = None
) -> DuvidaTrainingBatch:
    """
    """
    
    mgs, V_ds, x_ds, ys, weights, lt_masks, gt_masks = zip(*batch)
    bmg = DuvidaBatchMolGraph(mgs)
    training_batch = DuvidaTrainingBatch(
        bmg,
        None if V_ds[0] is None else torch.concat(V_ds),
        None if x_ds[0] is None else torch.stack(x_ds),
        None if ys[0] is None else torch.concat(ys).unsqueeze(1),
        weights,
        None if lt_masks[0] is None else lt_masks,
        None if gt_masks[0] is None else gt_masks,
    )
    if device is not None:
        training_batch.to(device)
    return training_batch


def _collate_training_batch_for_forward(
    batch: Iterable[Mapping],
    for_dataloader: bool = False,
    _in_key: str = "inputs",
    _out_key: str = "labels", 
    device: str = None
) -> TrainingBatch:
    new_batch = []
    for datum in batch:
        new_datum = {}
        if for_dataloader and _in_key in datum:
            _datum = datum[_in_key]
        else:
            _datum = datum
        for key, val in _datum.items():
            if isinstance(val, dict):
                new_val = MolGraph(**val)
            else:
                new_val = val
            new_datum[key] = new_val
        new_batch.append(Datum(**new_datum))
    collated = collate_batch(new_batch, device=device)
    if for_dataloader:
        return {
            _in_key: collated,
            _out_key: torch.concat([d[_out_key] for d in batch]).unsqueeze(1),
        }
    else:
        return collated



"""Classes to enable JSON serialization of preprocessing functions."""

from typing import Any, Dict, Iterable, Mapping, Union
from dataclasses import asdict, dataclass, field

import numpy as np

from .registry import FUNCTION_REGISTRY
from ...checkpoint_utils import save_json, _load_json


@dataclass
class Preprocessor:

    name: str
    input_column: str
    kwargs: Mapping[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        try:
            self.closure = FUNCTION_REGISTRY[self.name]
        except KeyError:
            raise ValueError(f"Function '{self.name}' is not registered.")
        self.hash = self._get_hash()
        self.hash_stub = self.hash[:7]
        self.output_column = self._get_output_column()
        self.function = self.closure(**self.kwargs)

    def _get_hash(self):
        from datasets.fingerprint import Hasher
        return Hasher.hash(self.to_dict())

    def _get_output_column(self):
        return f"__f__{self.name}__{self.hash_stub}__{self.input_column}"

    def to_dict(self) -> Dict[str, Any]:
        """Convert the function call to a JSON-serializable format."""
        return asdict(self)

    def to_file(self, filename: str) -> None:
        """Save to JSON."""
        return save_json(self.to_dict(), filename)

    @classmethod
    def from_dict(cls, data: Mapping[str, Union[str, Mapping]]) -> 'Preprocessor':
        """Reconstruct the function call from JSON."""
        return cls(**data)

    @classmethod
    def from_file(cls, filename: str) -> 'Preprocessor':
        """Load from JSON."""
        return cls.from_dict(_load_json(filename))

    @classmethod
    def show(cls) -> 'Preprocessor':
        """List registered functions."""
        return tuple(FUNCTION_REGISTRY)

    def __call__(self, inputs: Mapping[str, Iterable]) -> Dict[str, np.ndarray]:
        """Execute the function with stored parameters."""
        inputs[self.output_column] = self.function(inputs, self.input_column)
        return inputs



"""Approximate percentiles out-of-core."""

from typing import Dict, Iterable, Mapping, Union 

from functools import partial

from carabiner import print_err
from datasets import Dataset, IterableDataset, concatenate_datasets
import numpy as np
from tdigest import TDigest
from tqdm.auto import tqdm

from .utils import dataset_len
from ..datasets import to_dataset

def _tag_top(
    x: Mapping[str, Iterable],
    column: str,
    cutoff: float,
    output_column: str,
    delta: float = .01,
    reverse: bool = False
) -> Dict[str, Iterable]:
    out = []
    sign = -1 if reverse else 1
    for v in x[column]:
        if v >= cutoff + delta:
            out.append(sign)
        elif v <= cutoff - delta:
            out.append(-1 * sign)
        else:
            out.append(0)
    x[output_column] = out
    return x
    

def get_percentile(
    ds: Union[Dataset, IterableDataset],
    column: str,
    digest: TDigest,
    p: float,
    count: int,
    reverse: bool = False,
    delta: float = 1.,
    batch_size: int = 1024,
    cache: str = "./cache"
):
    if not reverse:
        cutoff = digest.percentile(100. - p)
    else:
        cutoff = digest.percentile(p)
    k = np.ceil(count * p / 100.).astype(int)
    print_err(f"Approx cutoff={cutoff:.4f}; need top {k} of {count} examples")

    top_col = f"{column}_top_{int(p)}_pc"
    if isinstance(ds, Dataset):
        desc = {"desc": "Tagging definitelys and maybes"}
    else:
        desc = {}
    ds = ds.map(
        _tag_top,
        fn_kwargs={
            "column": column,
            "cutoff": cutoff,
            "delta": delta,
            "output_column": top_col,
            "reverse": reverse,
        },
        batched=True,
        batch_size=batch_size,
        **desc,
    )
    
    ds_tfm = {}
    desc = "Separating definitelys from maybes ({})"
    for v in (1, -1, 0):
        if isinstance(ds, Dataset):
            _desc = {"desc": desc.format(v)}
        else:
            _desc = {}
        ds_tfm[v] = ds.filter(lambda x: x[top_col] == v, **_desc)
    
    n_definitely_true = dataset_len(ds_tfm[1])
    n_required_from_maybe = k - n_definitely_true
    
    if isinstance(ds, Dataset):
        desc = {"desc": "Tagging borderline"}
    else:
        desc = {}
    to_dataset_p = partial(to_dataset, batch_size=batch_size, cache=cache)
    maybes = to_dataset_p(ds_tfm[0], nrows=dataset_len(ds_tfm[0])).sort(column, reverse=reverse)

    print_err(f"Confidently tagged {n_definitely_true} {column=}, {p=}; need {n_required_from_maybe} / {maybes.num_rows} to make {k} total.")
    maybes_to_true = maybes.take(n_required_from_maybe).map(lambda x: {top_col: 1}, desc="Tagging maybes as True")
    try:
        maybes_to_false = maybes.skip(n_required_from_maybe).map(lambda x: {top_col: -1}, desc="Tagging maybes as False")
    except IndexError:
        maybes_resolved = maybes_to_true
    else:
        maybes_resolved = concatenate_datasets([
            maybes_to_true,
            maybes_to_false,
        ])
    final_ds = concatenate_datasets([
        to_dataset_p(ds_tfm[1], nrows=dataset_len(ds_tfm[1])), 
        maybes_resolved,
        to_dataset_p(ds_tfm[-1], nrows=dataset_len(ds_tfm[-1])),
    ]).map(
        lambda x: {top_col: [v == 1 for v in x[top_col]]}, 
        desc="Polishing percentile column",
        batched=True,
        batch_size=batch_size,
    )

    return final_ds


def percentiles(
    ds: Union[Dataset, IterableDataset],
    q: Mapping[str, Iterable[float]],
    compression: int = 500,
    delta: float = .01,
    reverse: bool = False,
    batch_size: int = 1024,
    cache: str = "./cache"
):
    digests = {
        key: TDigest(K=compression) for key in q
    }
    input_ds = ds
    for i, example in enumerate(tqdm(ds.iter(batch_size=batch_size), desc="Building quantile sketch")):
        # print(example)
        for key, digest in digests.items():
            # print(example[key])
            digest.batch_update(example[key])
    count = i * batch_size + len(example[key])

    for key, percentiles in q.items():
        for percentile in percentiles:
            ds = get_percentile(
                ds=ds,
                column=key,
                digest=digests[key],
                p=percentile,
                count=count,
                delta=delta,
                batch_size=batch_size,
                reverse=reverse,
                cache=cache,
            )
    if isinstance(input_ds, Dataset):
        return ds
    else:
        return ds.to_iterable_dataset()